"""This file contains tools for making property maps visualised on sphere"""

import abc
import os
import shutil
from pathlib import Path

import numpy as np
from numpy import linalg as lin
import pandas as pd

from hikari.dataframes import HklFrame, LstFrame
from hikari.symmetry import SG, Group
from hikari.utility import make_abspath, weighted_quantile, sph2cart, cart2sph,\
    Interval, artist_factory


class AngularPropertyExplorerFactory:
    """A factory method for creating angular property explorers."""
    def __init__(self):
        self._explorers = {}

    def register(self, prop, explorer):
        self._explorers[prop] = explorer

    def create(self, prop, **kwargs):
        explorer = self._explorers.get(prop)
        if not explorer:
            raise ValueError(f'Explorer for {prop} has not been registered!')
        return explorer(**kwargs)


class AngularPropertyExplorer:
    """
    An abstract base class for objects handling analysing parameters as
    a function of crystal orientation in a Diamond Anvil Cell.
    In order to generate a map of desired property,
    the following methods must be executed in order:

    - set_path
    - set_sample
    - set_output
    - explore
    """

    hkl_is_read_not_generated: bool
    property_name: str
    property_theoretical_limits: Interval

    GNUPLOT_INPUT_EXTENSION = '.gnu'
    GNUPLOT_OUTPUT_EXTENSION = '.pnG'
    HISTOGRAM_EXTENSION = '.his'
    HKL_EXTENSION = '.hkl'
    LISTING_EXTENSION = '.lst'
    MATPLOTLIB_EXTENSION = '.png'
    MESH_EXTENSION = '.dat'

    POLAR_LIMIT_DICT = {Group.System.triclinic: Interval(0, 180),
                        Group.System.monoclinic: Interval(0, 180),
                        Group.System.orthorhombic: Interval(0, 90),
                        Group.System.tetragonal: Interval(0, 90),
                        Group.System.trigonal: Interval(0, 90),
                        Group.System.hexagonal: Interval(0, 90),
                        Group.System.cubic: Interval(0, 90)}

    AZIMUTH_LIMIT_DICT = {Group.System.triclinic: Interval(-45, 135),
                          Group.System.monoclinic: Interval(0, 90),
                          Group.System.orthorhombic: Interval(0, 90),
                          Group.System.tetragonal: Interval(0, 90),
                          Group.System.trigonal: Interval(0, 120),
                          Group.System.hexagonal: Interval(0, 120),
                          Group.System.cubic: Interval(0, 90)}

    def __init__(self):
        self._path = ''
        self._sg = Group()
        self._pg = Group()
        self._lg = Group()
        self._th_limits = Interval(0, 0)
        self._ph_limits = Interval(0, 0)
        self.axis = ''
        self.hkl_frame = HklFrame()
        self.opening_angle = 0.0
        self.orientation = None
        self.resolution = 0.0
        self.axis = ''
        self.fix_scale = False
        self.histogram = False
        self.output_quality = 1
        self.data_dict = {'th': [], 'ph': [], 'potency': [], 'reflns': [],
                          'R1': [], 'weight': []}

    def set_up(self, a, b, c, al, be, ga, space_group, wavelength, axis,
               opening_angle, orientation, resolution,
               path, fix_scale, histogram, output_quality):
        self.opening_angle = opening_angle
        self.orientation = None if orientation is None \
            else np.array(orientation)
        self.resolution = resolution
        self.path = path
        self.fix_scale = fix_scale
        self.histogram = histogram
        self.output_quality = output_quality
        self.sg = SG[space_group]
        self.hkl_frame.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
        self.hkl_frame.la = wavelength
        self.axis = axis
        if self.hkl_is_read_not_generated:
            self._read_hkl_frame(hkl_format='shelx_4')
        else:
            self._make_hkl_frame()
        self.hkl_frame.find_equivalents(point_group=self.pg)
        total_unique = self.hkl_frame.table['equiv'].nunique()
        if total_unique == 0:
            raise KeyError('Specified part of reciprocal space has zero nodes')

    @abc.abstractmethod
    def explore(self):
        """
        Main interface method which, apart from the setters, is run within
        the script itself and generates data and figures using class methods
        """
        pass

    @property
    def prop_limits(self):
        if not self.fix_scale and self.data_dict[self.property_name]:
            _prop_limits = Interval(min(self.data_dict[self.property_name]),
                                    max(self.data_dict[self.property_name]))
        else:
            _prop_limits = self.property_theoretical_limits
        return _prop_limits

    def write_hist_file(self):
        hist_bins, hist_edges = np.histogram(
            self.data_dict[self.property_name], density=True,
            weights=self.data_dict['weight'], bins=32, range=self.prop_limits)
        hist_bins = hist_bins / sum(hist_bins)
        with open(self.path + self.HISTOGRAM_EXTENSION, 'w+') as h:
            h.write('#   from      to   prob.\n')
            for _f, _t, _p in zip(hist_edges[:-1], hist_edges[1:], hist_bins):
                h.write(f'{_f:8.5f}{_t:8.5f}{_p:8.5f}\n')

    def _make_hkl_frame(self):
        """Make ball or axis of hkl which will be cut in further steps"""
        f = self.hkl_frame
        f.fill(radius=min(self.hkl_frame.r_lim, self.resolution))
        if self.axis in {'x'}:
            f.table = f.table.loc[f.table['k'].eq(0) & f.table['l'].eq(0)]
        elif self.axis in {'y'}:
            f.table = f.table.loc[f.table['h'].eq(0) & f.table['l'].eq(0)]
        elif self.axis in {'z'}:
            f.table = f.table.loc[f.table['h'].eq(0) & f.table['k'].eq(0)]
        elif self.axis in {'xy'}:
            f.table = f.table.loc[f.table['l'].eq(0)]
        elif self.axis in {'xz'}:
            f.table = f.table.loc[f.table['k'].eq(0)]
        elif self.axis in {'yz'}:
            f.table = f.table.loc[f.table['h'].eq(0)]
        if self.axis in {'x', 'y', 'z', 'xy', 'xz', 'yz'}:
            f.transform([o.tf for o in self.pg.operations])
        f.extinct(self.sg)
        return f

    def _read_hkl_frame(self, hkl_format):
        """Read reflections of hkl which will be cut in further steps"""
        f = self.hkl_frame
        f.read(hkl_path=self.path + self.HKL_EXTENSION, hkl_format=hkl_format)
        f.trim(limit=min(f.r_lim, self.resolution))
        f.find_equivalents(point_group=self.lg)
        return f

    @property
    def th_limits(self):
        """Interval range of polar angle where property will be calculated"""
        return self._th_limits

    @property
    def ph_limits(self):
        """Interval range of azimuth angle where property will be calculated"""
        return self._ph_limits

    def _draw_map(self, artist, extension):
        a = artist
        a.x_axis = self.hkl_frame.a_w / lin.norm(self.hkl_frame.a_w)
        a.y_axis = self.hkl_frame.b_w / lin.norm(self.hkl_frame.b_w)
        a.z_axis = self.hkl_frame.c_w / lin.norm(self.hkl_frame.c_w)
        a.focus = self.focus_cartesian
        a.heat_limits = self.prop_limits
        a.heat_palette = self.axis
        a.histogram = self.histogram
        a.polar_limits = self.th_limits
        a.azimuth_limits = self.ph_limits
        a.plot(self.path + extension)

    def draw_matplotlib_map(self):
        a = artist_factory.create('matplotlib_angular_heatmap_artist')
        self._draw_map(artist=a, extension=self.MATPLOTLIB_EXTENSION)

    def draw_gnuplot_map(self):
        a = artist_factory.create('gnuplot_angular_heatmap_artist')
        self._draw_map(artist=a, extension=self.GNUPLOT_OUTPUT_EXTENSION)

    @property
    def angle_res(self):
        if self.output_quality not in {1, 2, 3, 4, 5}:
            raise KeyError('output_quality should be 1, 2, 3, 4 or 5')
        return {1: 15, 2: 10, 3: 5, 4: 2, 5: 1}[self.output_quality]

    @property
    def path(self):
        """
        Provides path = directory + stem to the workspace. They are common to
        all input and output files, the extensions are specified as class var.
        """
        return self._path

    @path.setter
    def path(self, path):
        abs_path = Path(make_abspath(path))
        dir_, stem, ext = abs_path.parent, abs_path.stem, abs_path.suffix
        self._path = str(Path().joinpath(dir_, stem))

    @property
    def sg(self):
        return self._sg

    @sg.setter
    def sg(self, value):
        _pg = value.reciprocate()
        _sg_system = value.system
        self._sg = value
        self._pg = _pg
        self._lg = _pg.lauefy()
        self._th_limits = self.POLAR_LIMIT_DICT[_sg_system]
        self._ph_limits = self.AZIMUTH_LIMIT_DICT[_sg_system]

    @property
    def pg(self):
        return self._pg

    @property
    def lg(self):
        return self._lg

    @property
    def focus_cartesian(self):
        _focus = []
        if self.orientation is not None:
            if len(self.orientation.shape) == 1:
                a = self.orientation
            elif len(self.orientation.shape) == 2:
                a = lin.inv(self.orientation) @ np.array((1, 0, 0))
            else:
                raise ValueError(f'Unknown orientation: {self.orientation}')
            for op in self.lg.operations:
                v = self.hkl_frame.A_r.T @ op.tf @ a
                c = np.rad2deg(cart2sph(*v))
                if c[1] in self.th_limits and c[2] in self.ph_limits:
                    _focus.append(v / lin.norm(v))
        return _focus

    @property
    def focus_spherical_closest(self):
        _focus = []
        if self.orientation is not None:
            for focus_vector in self.focus_cartesian:
                _, th, ph = np.rad2deg(cart2sph(*focus_vector))
                th_closest = min(self.th_range, key=lambda x: abs(x - th))
                ph_closest = min(self.ph_range, key=lambda x: abs(x - ph))
                _focus.append([1.0, th_closest, ph_closest])
        return _focus

    @property
    def prop_at_focus_closest(self):
        df = pd.DataFrame({'ph': self.data_dict['ph'],
                           'th': self.data_dict['th'],
                           'p': self.data_dict[self.property_name]})
        properties_ = []
        for _, th, ph in self.focus_spherical_closest:
            row = df.loc[(df['th'] == th) & (df['ph'] == ph)].iloc[0]
            properties_.append(row['p'])
        return properties_

    @property
    def th_range(self):
        return self.th_limits.arange(step=self.angle_res)

    @property
    def ph_range(self):
        return self.ph_limits.arange(step=self.angle_res)

    @property
    def th_comb(self):
        return self.th_limits.comb_with(self.ph_limits, step=self.angle_res)[0]

    @property
    def ph_comb(self):
        return self.th_limits.comb_with(self.ph_limits, step=self.angle_res)[1]

    @property
    def th_mesh(self):
        return self.th_limits.mesh_with(self.ph_limits, step=self.angle_res)[0]

    @property
    def ph_mesh(self):
        return self.th_limits.mesh_with(self.ph_limits, step=self.angle_res)[1]

    def orientation_weights(self, th, ph):
        """Calculate how much each point should contribute to distribution"""
        def sphere_cutout_area(th1, th2, ph_span):
            """Calculate sphere area in specified ph and th degree range.
            For exact math, see articles about spherical cap and sector."""
            return np.deg2rad(abs(ph_span)) * \
                   abs(np.cos(np.deg2rad(th1)) - np.cos(np.deg2rad(th2)))
        th_max = (th + self.angle_res / 2.).clip(max=self.th_limits[1])
        th_min = (th - self.angle_res / 2.).clip(min=self.th_limits[0])
        ph_max = (ph + self.angle_res / 2.).clip(max=self.ph_limits[1])
        ph_min = (ph - self.angle_res / 2.).clip(min=self.ph_limits[0])
        return sphere_cutout_area(th_min, th_max, ph_max-ph_min)

    @property
    def descriptive_statistics_string(self):
        min_p = min(self.data_dict[self.property_name])
        worst_index = self.data_dict[self.property_name].index(min_p)
        worst_th = self.data_dict['th'][worst_index]
        worst_ph = self.data_dict['ph'][worst_index]

        max_p = max(self.data_dict[self.property_name])
        best_index = self.data_dict[self.property_name].index(max_p)
        best_th = self.data_dict['th'][best_index]
        best_ph = self.data_dict['ph'][best_index]

        avg_p = np.average(self.data_dict[self.property_name],
                           weights=self.data_dict['weight'])
        q1, q2, q3 = weighted_quantile(self.data_dict[self.property_name],
                                       quantiles=[0.25, 0.50, 0.75],
                                       weights=self.data_dict['weight'])

        s = f'# descriptive statistics for {self.property_name}:\n' \
            f'# max ={max_p:8.5f} at th ={best_th :6.1f} ph ={best_ph :6.1f}\n'\
            f'# min ={min_p:8.5f} at th ={worst_th:6.1f} ph ={worst_ph:6.1f}\n'\
            f'# q_1 ={q1   :8.5f}\n' \
            f'# q_2 ={q2   :8.5f}\n' \
            f'# q_3 ={q3   :8.5f}\n' \
            f'# avg ={avg_p:8.5f}\n'
        return s

    @property
    def prop_at_focus_string(self):
        s = f'# value of {self.property_name} at focus points:\n'
        for focus_, prop_ in zip(self.focus_spherical_closest,
                                 self.prop_at_focus_closest):
            _, th, ph = focus_
            s += f'# val ={prop_:8.5f} at th ={th :6.1f} ph ={ph :6.1f}\n'
        return s + '\n'


class AngularPotencyExplorer(AngularPropertyExplorer):
    hkl_is_read_not_generated = False
    property_name = 'potency'
    property_theoretical_limits = Interval(0, 1)

    def explore(self):
        dat_path = self.path + self.MESH_EXTENSION
        lst_path = self.path + self.LISTING_EXTENSION

        potency_mesh = np.zeros_like(self.th_mesh, dtype=float)
        lst = open(lst_path, 'w+')
        lst.write('#     th      ph potency  reflns\n')
        vectors = sph2cart(r=np.ones_like(self.th_comb),
                           p=np.deg2rad(self.th_comb),
                           a=np.deg2rad(self.ph_comb)).T
        weights = self.orientation_weights(th=self.th_comb, ph=self.ph_comb)
        uniques = self.hkl_frame.dacs_count(self.opening_angle, vectors=vectors)
        total_unique = self.hkl_frame.table['equiv'].nunique('')

        for i, th in enumerate(self.th_range):
            for j, ph in enumerate(self.ph_range):
                unique = uniques[j * len(self.th_range) + i]
                weight = weights[j * len(self.th_range) + i]
                potency = unique / total_unique
                self.data_dict['th'].append(th)
                self.data_dict['ph'].append(ph)
                self.data_dict['potency'].append(potency)
                self.data_dict['reflns'].append(unique)
                self.data_dict['weight'].append(weight)
                lst.write(f'{th:8.0f}{ph:8.0f}{potency:8.5f}{unique:8d}\n')
                potency_mesh[j][i] = potency
            lst.write('\n')

        np.savetxt(dat_path, potency_mesh)
        if self.orientation is not None:
            lst.write(self.prop_at_focus_string)
        lst.write(self.descriptive_statistics_string)
        lst.close()


class AngularR1Explorer(AngularPropertyExplorer):

    hkl_is_read_not_generated = True
    property_name = 'R1'
    property_theoretical_limits = Interval(0, 1)

    def explore(self):
        dat_path = self.path + self.MESH_EXTENSION
        lst_path = self.path + self.LISTING_EXTENSION
        job_name = Path(self.path).stem

        r1_mesh = np.zeros_like(self.th_mesh, dtype=float)
        lst = open(lst_path, 'w+')
        lst.write('#     th      ph      R1  reflns\n')

        weights = self.orientation_weights(th=self.th_comb, ph=self.ph_comb)
        total_unique = self.hkl_frame.table['equiv'].nunique()

        for i, th in enumerate(self.th_range):
            for j, ph in enumerate(self.ph_range):
                subdir = self.path + f'_th{int(th+.1)}_ph{int(ph+.1)}'
                hkl_path2 = make_abspath(subdir, job_name+'.hkl')
                ins_path2 = make_abspath(subdir, job_name+'.ins')
                lst_path2 = make_abspath(subdir, job_name+'.lst')
                dir_path2 = make_abspath(subdir)
                Path(dir_path2).mkdir()
                shutil.copy(self.path + '.res', ins_path2)
                q = self.hkl_frame.copy()
                q.dac_trim(self.opening_angle,
                           sph2cart(1.0, np.deg2rad(th), np.deg2rad(ph)))
                q.write(hkl_path2, hkl_format='shelx_4')
                unique = q.table['equiv'].nunique()
                weight = weights[j * len(self.th_range) + i]
                os.system(f'cd {dir_path2}; shelxl {job_name}')
                r1 = LstFrame().read_r1(lst_path2)
                potency = unique / total_unique
                self.data_dict['th'].append(th)
                self.data_dict['ph'].append(ph)
                self.data_dict['potency'].append(potency)
                self.data_dict['R1'].append(r1)
                self.data_dict['weight'].append(weight)
                lst.write(f'{th:8.0f}{ph:8.0f}{r1:8.5}{potency:8.5f}\n')
                r1_mesh[j][i] = r1
            lst.write('\n')

        np.savetxt(dat_path, r1_mesh)
        if self.orientation is not None:
            lst.write(self.prop_at_focus_string)
        lst.write(self.descriptive_statistics_string)
        lst.close()


angular_property_explorer_factory = AngularPropertyExplorerFactory()
angular_property_explorer_factory.register(
    prop='potency', explorer=AngularPotencyExplorer)
angular_property_explorer_factory.register(
    prop='r1', explorer=AngularR1Explorer)




"""
This file contains basic tools to work with numpy objects used in the package.
"""

import re
import numpy as np


def str2array(s: str) -> np.ndarray:
    """
    Create a numpy int8 array using its compact representation saved as string.
    The input string might contain only digits and special characters.
    Every individual digit is treated as a new element, while the special
    characters induce special behaviour and include:

    - slash `/` - if present, divides main list into multiple sub-lists;
    - hyphen `-` - the next digit will be read as negative

    Please mind that the compact form used as this function's input is unfit
    to store floating point numbers or integers with absolute value above 9.

    :param s: input string containing only digits and special characters
    :return: a two-dimensional numpy array with integers
    """
    elements = [re.findall(r"-?\d", column) for column in s.split('/')]
    if len({len(element_column) for element_column in elements}) != 1:
        raise ValueError('All columns specified with "/" must have same length')
    return np.array(elements, dtype=np.int8)



"""
This file contains constants and functions containing purely mathematical
information / methods used in the package.
"""

import numpy as np
import random
import uncertainties
from typing import Union


def angle2rad(value):
    """
    Interpret a unit of given angle value and return this value in radians.
    Values in range between -3.15 and 3.15 are interpreted as given in radians.
    Values outside this range are interpreted as given in degrees.

    The function can be used to dynamically interpret and accept
    both radian and degrees and return the value in angles. Should you not
    need this functionality, please use numpy's rad2deg and deg2rad instead.

    :Example:

    >>> angle2rad(np.pi)
    3.141592653589793
    >>> angle2rad(180)
    3.141592653589793
    >>> angle2rad(2*np.pi)
    0.1096622711232151

    Please mind that in the last case the value given in radians was treated
    as given in degrees and wrongly recalculated to radians again.

    :param value: Angle value given in angles or radians
    :type value: float
    :return: Angle value expressed in radians
    :rtype: float
    """
    return value if -3.15 < value < 3.15 else value * np.pi / 180


def cart2sph(x, y, z):
    """
    Convert Cartesian coordinates x, y, z
    to conventional spherical coordinates r, p, a

    :param x: Cartesian coordinate or vector x
    :type x: float or np.ndarray
    :param y: Cartesian coordinate or vector y
    :type y: float or np.ndarray
    :param z: Cartesian coordinates or vector z
    :type z: float or np.ndarray
    :return: Spherical coordinates: radius, polar angle, and azimuth angle
    :rtype: np.ndarray
    """
    r = (x ** 2 + y ** 2 + z ** 2) ** 0.5
    p = np.arccos(z / r)
    a = np.arctan2(y, x)
    return np.array([r, p, a])


def sph2cart(r, p, a):
    """
    Convert conventional spherical coordinates r, p, a
    to Cartesian coordinates x, y, z

    :param r: Spherical coordinate or vector radius
    :type r: float or np.ndarray
    :param p: Spherical coordinate or vector polar angle
    :type p: float or np.ndarray
    :param a: Spherical coordinate or vector azimuth angle
    :type a: float or np.ndarray
    :return: Cartesian coordinates: x, y, and z
    :rtype: np.ndarray
    """
    x = r * np.cos(a) * np.sin(p)
    y = r * np.sin(a) * np.sin(p)
    z = r * np.cos(p)
    return np.array([x, y, z])


def det3x3(matrix):
    """
    Calculate a determinant of a 3x3 matrix. Should be usually substituted
    by `numpy.linalg.det`, but is indispensable for matrices with uncertainties.

    :param matrix: 3x3 array/matrix which allows for 2d indexing
    :type matrix: numpy.ndarray or uncertainties.unumpy.matrix
    :return: Determinant of the matrix
    :rtype: int or float or uncertainties.core.Variable
    """
    m11, m12, m13 = matrix[0, 0], matrix[1, 0], matrix[2, 0]
    m21, m22, m23 = matrix[0, 1], matrix[1, 1], matrix[2, 1]
    m31, m32, m33 = matrix[0, 2], matrix[1, 2], matrix[2, 2]
    return m11 * m22 * m33 + m12 * m23 * m31 + m13 * m21 * m32 \
        - m13 * m22 * m31 - m12 * m21 * m33 - m11 * m23 * m32


def fibonacci_sphere(samples=1, seed=1337):
    """
    Return a 3D cartesian coordinates of *samples* number of points
    evenly distributed on a surface of a unit sphere at (0, 0, 0).

    The algorithm utilised in this function gives a uniform distribution on the
    sphere. Contrary to an uniform distribution on coordinates, this algorithm
    does not favour points close to poles. For more details see `this article
    <http://extremelearning.com.au/evenly-distributing-points-on-a-sphere/>`_.

    The function is written so that it does always return the same set of points
    in order to ensure reproducibility of other methods. A list of other points
    can be obtained by changing (or even randomising) the value of a *seed*.

    :Example:

    >>> fibonacci_sphere(4)
    array([[ 0.62069, -0.75, -0.22857],
           [-0.44395, -0.25,  0.86047],
           [ 0.41275,  0.25,  0.87587],
           [-0.61208,  0.75, -0.25072]])
    >>> fibonacci_sphere(4)
    array([[ 0.62069, -0.75, -0.22857],
           [-0.44395, -0.25,  0.86047],
           [ 0.41275,  0.25,  0.87587],
           [-0.61208,  0.75, -0.25072]])
    >>> fibonacci_sphere(4, seed=420)
    array([[ 0.64040, -0.75,  0.16550],
           [-0.85489, -0.25,  0.45460],
           [ 0.32329,  0.25, -0.91268],
           [ 0.25831,  0.75,  0.60892]])

    :param samples: Number of points to be generated.
    :type samples: int
    :param seed: A seed value used once to slightly randomise point position.
    :type seed: any
    :return: (x,y,z) arrays representing points' cartesian coordinates.
    :rtype: np.array
    """
    random.seed(seed)
    rnd = random.random() * samples
    offset = 2. / samples
    increment = np.pi * (3. - np.sqrt(5.))
    i = np.arange(0, samples)
    y = ((i * offset) - 1) + (offset / 2)
    r = np.sqrt(1 - pow(y, 2))
    phi = ((i + rnd) % samples) * increment
    x = np.cos(phi) * r
    z = np.sin(phi) * r
    return np.vstack([x, y, z]).T


def euler_rodrigues_matrix(
        a: Union[int, float, uncertainties.UFloat],
        b: Union[int, float, uncertainties.UFloat],
        c: Union[int, float, uncertainties.UFloat],
        d: Union[int, float, uncertainties.UFloat]) -> np.ndarray:
    """
    Return a rotation matrix based on a Euler-Rodrigues parametrisation. For
    details, see https://en.wikipedia.org/wiki/Euler%E2%80%93Rodrigues_formula.

    :param a: Euler-Rodrigues parameter *a*
    :param b: Euler-Rodrigues parameter *b*
    :param c: Euler-Rodrigues parameter *c*
    :param d: Euler-Rodrigues parameter *d*
    :return: 3x3 matrix describing rotation
    """
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


def rotation_from(from_, to):
    """
    Return matrix associated with rotation of vector `from_` onto `to`.

    :param from_: A 3D vector which will have been rotated onto `to`.
    :type from_: np.ndarray
    :param to: A 3D vector onto which `from` will have been rotated.
    :type to: np.ndarray
    :return: Matrix associated with rotation of `from` onto `to`.
    :rtype: np.ndarray
    """
    if from_.size != 3:
        raise IndexError(f'Size of `from_` should be 3, but is {from_.size}')
    if to.size != 3:
        raise IndexError(f'Size of `to` should be 3, but is {to.size}')
    f = from_ / np.linalg.norm(from_)
    t = to / np.linalg.norm(to)
    axis = np.cross(f, t)
    if np.isclose(sum(axis), 0.):
        if np.allclose(f, t):
            return np.eye(3)
        else:
            e1, e2 = np.array([1, 0, 0]), np.array([0, 1, 0])
            e = np.cross(e2, f) if np.isclose(sum(np.cross(e1, f)), 0.) \
                else np.cross(e1, f)
            return rotation_around(e, by=np.pi)
    else:
        axis = axis / np.linalg.norm(axis)
        angle = np.arcsin(np.linalg.norm(np.cross(f, t)))
        return euler_rodrigues_matrix(np.cos(angle / 2.),
                                      *(-axis * np.sin(angle / 2.)))


def rotation_around(axis, by):
    """
    Return matrix associated with counterclockwise rotation about `axis` through
    `by` radians. See Euler-Rodrigues form.,https://stackoverflow.com/q/6802577/

    :param axis: A 3D vector about which rotation is performed
    :type axis: np.ndarray
    :param by: Angle of rotation in radians
    :type by: float
    :return: Matrix associated with rotation around `axis` through `by`.
    :rtype: np.ndarray
    """
    if axis.size != 3:
        raise IndexError(f'Size of `axis` should be 3, but is {axis.size}')
    axis = axis / np.linalg.norm(axis)
    return euler_rodrigues_matrix(np.cos(by / 2.0), *(-axis * np.sin(by / 2.0)))


def weighted_quantile(values, quantiles, weights=None):
    """
    A function to approximate quantiles based on many weighted points. Quantiles
    are derived exactly using method 7 of H&F from an unweighted distribution,
    otherwise they are approximated by comparing two opposing interpolations.

    Let A: {1, 2, 2, 2, 3} and B: {1, 2(x3), 3} be seemingly equivalent sets
    where (x3) denotes triple weight of the "2". Let q[n] denote quantiles.
    For set A, q[1/4](A) = q[3/4](A) = 2. However, for set B we need to
    extrapolate and read quantile from PDF, obtained via linear interpolation.
    As a result, q[1/4](B) = 4/3 < q[3/4](B) = 2 since f(0) = 1 and f(3/4) = 2
    at f(1/4) and the results are always underestimated.

    To combat this effect, we can repeat the process for -B: {-3, -2(x3), -1}.
    Here, q[1/4](-B) = -8/3 and q[3/4](-B) = -2. As a result, the quantile
    q[n](A) is described a bit better using formula (q[n](B) - q[1-n](-B)) / 2.

    :param values: data to be evaluated
    :type values: tuple or list or np.ndarray
    :param quantiles: iterable containing desired quartiles between 0 and 1
    :type quantiles: tuple or list or np.ndarray
    :param weights: weights of data in the same order
    :type weights: tuple or list or np.ndarray
    :return: array containing approximated quartiles
    :rtype: np.ndarray
    """

    values = np.array(values)
    quantiles = np.array(quantiles)
    w = np.ones_like(values) if weights is None else np.array(weights)
    if np.any(quantiles < 0) or np.any(quantiles > 1):
        raise ValueError(f'quantiles ({quantiles}) should be in range 0 to 1')

    sorter = np.argsort(values)
    values = values[sorter]
    w = w[sorter]

    weighted_cumsum_0to1 = (np.cumsum(w) - w[0]) / np.sum(w[1:])
    weighted_cumsum_1to0 = (np.cumsum(np.flip(w)) - w[-1]) / np.sum(w[:-1])
    results_0to1 = np.interp(quantiles, weighted_cumsum_0to1, values)
    results_1to0 = np.interp(1-quantiles, weighted_cumsum_1to0, np.flip(values))
    return (results_0to1 + results_1to0) / 2.0



"""
This sub-module contains scripts made to work mainly with .hkl files,
containing information from single crystal diffraction experiments.
Due to authors interest in the subjects, majority of them aim to analyse
or model data from high-pressure experiments,
which utilise DAC - diamond anvil cell.
"""

from hikari.dataframes import HklFrame
from hikari.symmetry import SG
from hikari.utility import cubespace, make_abspath
import numpy as np


def completeness_statistics(a, b, c, al, be, ga,
                            space_group='P-1',
                            input_path='shelx.hkl',
                            input_format='shelx_4',
                            input_wavelength='CuKa'):
    """
    For a given experimental .hkl file calculate basic completeness statistics
    in ten resolution shells of equal-volume.

    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *alpha* in degrees.
    :type be: float
    :param ga: Unit cell parameter *alpha* in degrees.
    :type ga: float
    :param space_group: Short Hermann-Mauguin name or index of space group.
        For details see table in hikari.symmetry.space_groups.
    :type space_group: str or int
    :param input_path: Path to the input .hkl file.
    :type input_path: str
    :param input_format: Format of the .hkl file. For reference
        see :meth:`hikari.dataframes.HklFrame.interpret_hkl_format`.
    :type input_format: str or dict
    :param input_wavelength: Wavelength of radiation utilised in experiment.
    :type input_wavelength: float or str
    :return: None
    :rtype: None
    """
    p = HklFrame()
    p.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
    p.la = input_wavelength
    p.read(make_abspath(input_path), input_format)
    print(p.stats(space_group=SG[space_group]))


def dac_statistics(a, b, c, al, be, ga,
                   space_group='P1',
                   opening_angle=35.0,
                   orientation=((1, 0, 0), (0, 1, 0), (0, 0, 1)),
                   input_path='shelx.hkl',
                   input_format=4,
                   input_wavelength='CuKa',
                   resolution=None):
    """
    For a given experimental .hkl file calculate number of experimentally found
    and theoretically possible reflections up to a given resolution.

    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *alpha* in degrees.
    :type be: float
    :param ga: Unit cell parameter *alpha* in degrees.
    :type ga: float
    :param space_group: Short Hermann-Mauguin name or index of space group.
        For details see table in hikari.symmetry.space_groups.
    :type space_group: str or int
    :param opening_angle: Value of single opening angle as defined
        in :py:meth:`hikari.dataframes.HklFrame.dac`.
    :type opening_angle: float
    :param orientation: Crystal orientation as defined
        in :py:class:`hikari.dataframes.BaseFrame`
    :type orientation: tuple or numpy.array
    :param input_path: Path to the input .hkl file.
    :type input_path: str
    :param input_format: Format of the .hkl file. For reference
        see :py:meth:`hikari.dataframes.HklFrame.interpret_hkl_format`.
    :type input_format: int or str or dict
    :param input_wavelength: Wavelength of radiation utilised in experiment.
    :type input_wavelength: float or str
    :param resolution: If given, calculate statistics only up to this value.
        Please provide it as a distance in rec. space (twice the resolution in A-1).
    :type resolution: float
    :return: None
    """

    sg = SG[space_group]
    pg = sg.reciprocate()

    p = HklFrame()
    p.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
    p.la = input_wavelength
    p.read(make_abspath(input_path), input_format)
    p.orientation = np.array(orientation)

    resolution = p.r_lim if resolution is None else resolution
    p.trim(limit=resolution)
    p.merge(point_group=pg)

    q = p.copy()
    q.fill(radius=resolution)
    q.dac_trim(opening_angle=opening_angle)

    # uncomment and fill this if you have 2 crystals in different orientations
    # q2 = p.copy()
    # q2.fill(radius=resolution)
    # q2.orientation = np.array(())
    # q2.dac(opening_angle=opening_angle)
    # q = q + q2

    q.merge(point_group=pg)
    q.extinct(space_group=sg)

    b = p.copy()
    b.fill(radius=resolution)
    b.extinct(space_group=sg)
    b.merge(point_group=pg)

    r_max = resolution
    print('radius    experimnt theoryDAC theorBall DAC-Cplt  Ball-Cplt ')
    print('range     unique    unique    unique    exp/DAC   exp/Ball  ')
    for rad in reversed(cubespace(0, r_max, 10, include_start=False)):
        p.trim(rad)
        q.trim(rad)
        b.trim(rad)
        p_eq = p.table['equiv'].nunique()
        q_eq = q.table['equiv'].nunique()
        b_eq = b.table['equiv'].nunique()
        print(' {:9f}'.format(rad) + ' {:9d}'.format(p_eq) +
              ' {:9d}'.format(q_eq) + ' {:9d}'.format(b_eq) +
              ' {:9f}'.format(p_eq / q_eq) + ' {:9f}'.format(p_eq / b_eq))


def reformat_hkl(input_path: str,
                 input_format: str,
                 output_path: str,
                 output_format: str):
    """
    :param input_path: Relative or absolute path to the input .hkl file.
    :param input_format: Format of the .hkl file. For reference
        see :attr:`hikari.dataframes.HklIo.format`.
    :param output_path: Relative or absolute path to the output .hkl file.
    :param output_format: Format of the .hkl file. For reference
        see :attr:`hikari.dataframes.HklIo.format`.
    :return:
    :rtype:
    """
    h = HklFrame()
    absolute_input_path = make_abspath(input_path)
    absolute_output_path = make_abspath(output_path)
    h.read(hkl_path=absolute_input_path, hkl_format=input_format)
    h.write(hkl_path=absolute_output_path, hkl_format=output_format)


def simulate_dac(a, b, c, al, be, ga,
                 opening_angle=35,
                 orientation=((1, 0, 0), (0, 1, 0), (0, 0, 1)),
                 vector=None,
                 resolution=None,
                 input_path='shelx.hkl',
                 input_format=4,
                 input_wavelength='CuKa',
                 output_path='output.hkl',
                 output_format=4):
    """
    For a given experimental .hkl file simulate a lack of completeness
    caused by a presence of high-pressure diamond anvil cell.

    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *alpha* in degrees.
    :type be: float
    :param ga: Unit cell parameter *alpha* in degrees.
    :type ga: float
    :param opening_angle: Value of single opening angle as defined in
        :meth:`hikari.dataframes.HklFrame.dac`.
    :type opening_angle: float
    :param orientation: Crystal orientation as defined in
        :class:`hikari.dataframes.BaseFrame`
    :type orientation: tuple or numpy.array
    :param vector: If given, overwrite orientation to provide information
        about crystal placement in dac,
        as defined in :meth:`hikari.dataframes.HklFrame.dac`.
    :type vector: tuple
    :param resolution: If given, additionally limit data resolution to given
        value. Please provide the resolution as a distance from the origin
        in reciprocal space (twice the resolution in reciprocal angstrom).
    :type resolution: float
    :param input_path: Path to the input .hkl file.
    :type input_path: str
    :param input_format: Format of the input .hkl file. For reference see
        :meth:`hikari.dataframes.HklFrame.interpret_hkl_format`.
    :type input_format: int or str or dict
    :param input_wavelength: Wavelength of radiation utilised in experiment.
    :type input_wavelength: float or str
    :param output_path: Path to the output .hkl file.
    :type output_path: str
    :param output_format: Format of the input .hkl file. For reference see
        :meth:`hikari.dataframes.HklFrame.interpret_hkl_format`.
    :type output_format: int or str or dict
    :return: None
    """
    p = HklFrame()
    p.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
    p.la = input_wavelength
    p.read(make_abspath(input_path), input_format)
    p.orientation = np.array(orientation)
    if not(resolution is None):
        p.trim(resolution)
    p.dac_trim(opening_angle=opening_angle, vector=vector)
    p.write(hkl_path=make_abspath(output_path), hkl_format=output_format)


if __name__ == '__main__':
    completeness_statistics(5.641087, 5.641087, 5.641087, 90, 90, 90,
                            space_group='Fm-3m', input_path='~/_/NaCl/NaCl.hkl',
                            input_format='shelx_4', input_wavelength='CuKa')



class Singleton(type):
    """
    A metaclass based on https://stackoverflow.com/q/6760685/.
    Declare a singleton class using `MyClass(BaseClass, metaclass=Singleton)`
    """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot
from numpy import linalg as lin

from hikari.dataframes import HklFrame
from hikari.symmetry import SG, Group
from hikari.utility import make_abspath, fibonacci_sphere, rotation_around
from hikari.scripts.angular_explorer import angular_property_explorer_factory


def potency_map(a, b, c, al, be, ga,
                space_group='P1',
                axis='',
                fix_scale=False,
                histogram=True,
                opening_angle=35,
                orientation=None,
                path='~/sortav.lst',
                output_quality=3,
                resolution=1.2,
                wavelength='MoKa'):
    r"""
    Calculate and draw a potency map for a given crystal in diamond anvil cell
    (DAC) with a given opening angle, as a function of crystal orientation.
    For details see `this paper <https://doi.org/10.1107/S2052252521009532>`_.

    The script accepts unit cell & space group information, and predicts max
    completeness of fully merged data for investigated crystal in said group.
    Results are logged into text files and drawn with gnuplot or matplotlib.

    Potency is calculated and visualised on a unit-sphere orientation heatmap.
    Each point **p** is associated with a certain crystal orientation in DAC,
    such that vector **v** from **0** to **p** acts as a symmetry axis for the
    dac-accessible volume traced in the reciprocal space up to `resolution`.
    Red / green / blue vectors represent crystallographic directions **X\***,
    **Y\*** and **Z\***, respectively. Since the distribution has some inherent
    symmetry, only a representative part of sphere (usually an octant) is shown.

    As an example, let's assume a orthorhombic cell with *a* = *b* = *c* = 10
    and laue group "mmm". Running the script and generating the potency the map
    yields the lowest values close to **X\***, **Y\*** and **Z\*** vector,
    while the highest values are observed between those vectors.
    Placing the crystal on its [100] face inside the dac will cause the
    dac-accessible plane to be placed perpendicularly to (100) direction
    in reciprocal space. Since the values close to **X\*** ((100) direction)
    are low, such a placement will allow us to collect data with low coverage.
    On the other hand, placing the crystal on its [111] face will cause the
    dac-accessible plane to be placed perpendicularly to (100) direction
    in reciprocal space. As on the sphere the values tend to rise the further
    we are from of **X\***, **Y\*** and **Z\***, we expect this crystal
    orientation to warrant a high completeness of collected data.

    The potency is calculated as a ratio of the number of unique reflections
    inside the DAC-accessible space to the number of unique reflections
    inside a reference sphere of a radius equal to `resolution`.
    Visualisation is performed using an extended rainbow heatmap,
    which utilises a wide color range to emphasize even small differences.
    Dy default, the color scale is dynamic and adapts to the range of calculated
    potency, but it can be fixed to 0-100% range using `fix_scale=True`.

    Since the orientation is given in spherical coordinates, the exact positions
    of individual points is given using *theta* and *phi* angles instead of
    crystallographic coordinates. The *theta* and *phi* angles here follow the
    physical definition (ISO 80000-2:2019). For a given DAC-axis vector **v**:

    - *theta* is the azimuth angle found between vector **Z\*** and **v**.
      It can assume values between zero and *pi* (*pi/2* in one octant).

    - *phi* is the rotational angle denoting rotation of **v** around **Z\***.
      It can assume values between zero and *2 pi* (*pi/2* in one octant).

    Finally it must be noted that higher potency setting is not always better.
    For example, for opening angle below 45 degrees, most potent orientation in
    laue class mmm renders all 0kl, h0l, hk0 reflections inaccessible.
    Potency map should be consulted before a high-pressure experiment, but it
    should not be treated as an universal quality indicator of given set-up.

    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *alpha* in degrees.
    :type be: float
    :param ga: Unit cell parameter *alpha* in degrees.
    :type ga: float
    :param space_group: Short Hermann-Mauguin name or index of space group.
        For details see :py:mod:`hikari.symmetry.space_groups`.
    :type space_group: str or int
    :param axis: domain to calculate potency in. Accepts 'x'/ 'y'/ 'z' for h00/
        0k0/ 00l, 'xy'/'xz'/'yz' for hk0/ h0l/ 0kl, or '' for all reflections.
    :type axis: string
    :param fix_scale: If true, the colour scheme will fix to 0 - 100% range.
    :type fix_scale: bool
    :param histogram: If true, potency distribution will be plotted as histogram
    :type histogram: bool
    :param opening_angle: Value of single opening angle as defined in
        :meth:`hikari.dataframes.HklFrame.dac`.
    :type opening_angle: float
    :param orientation: either a cif-style 3x3 matrix of crystal orientation or
        a 3-length array with a diamond-perpendicular face to be marked on a map
    :type orientation: np.ndarray
    :param path: Path to a file where script is to be run. Extension is ignored,
        but file name will and must be the same for all input and output files.
    :type path: str
    :param output_quality: Density of individual orientations to be considered.
        Should be in range from 1 (every 15 degrees) to 5 (every 1 degree).
    :type output_quality: int
    :param resolution: Upper limit of reflection resolution, given as a distance
        from zero to node in reciprocal space (one over plane spacing) in A-1.
    :type resolution: float
    :param wavelength: Wavelength of radiation to be simulated.
    :type wavelength: float or str
    :return: None
    :rtype: None
    """
    kwargs = locals()
    ape = angular_property_explorer_factory.create(prop='potency')
    ape.set_up(**kwargs)
    ape.explore()
    ape.write_hist_file()
    ape.draw_matplotlib_map()
    ape.draw_gnuplot_map()


def potency_vs_dac_opening_angle(output_path='~/output.txt',
                                 precision=91,
                                 resolution=1.2,
                                 wavelength='MoKa',
                                 theta=None):
    """
    Calculate potency in P1 space group as a function of DAC opening angle,
    assuming certain resolution and wavelength used.

    :param output_path: Path of created file containing calculated data.
    :type output_path: str
    :param precision: Number of probed opening angles between 0 and 90 degrees.
    :type precision: int
    :param resolution: Maximum distance from the origin in reciprocal space to
        reflection (twice the resolution in reciprocal angstrom). Default 1.2.
    :type resolution: float
    :param wavelength: Wavelength of radiation to be simulated.
    :type wavelength: float or str
    :param theta: If given, use max theta in degrees (instead of radius in A-1).
    :type theta: float
    :return: None
    """

    def _make_reference_ball():
        hkl_frame = HklFrame()
        hkl_frame.la = wavelength
        res = hkl_frame.r_lim * np.sin(np.deg2rad(theta)) \
            if theta is not None else resolution
        side = 10 * precision**(1/3) / res  # adapt to res&la
        hkl_frame.edit_cell(a=side, b=side, c=side, al=90, be=90, ga=90)
        hkl_frame.fill(radius=res)
        hkl_frame.find_equivalents()
        return hkl_frame

    p = _make_reference_ball()
    v = fibonacci_sphere(100)
    total = len(p)
    angles = np.linspace(start=90, stop=0, num=precision)
    with open(make_abspath(output_path), 'w', buffering=1) as out:
        out.write('#     oa potency\n')
        for a in angles:  # for one random vector v
            c = p.dacs_count(opening_angle=a, vectors=v)
            out.write(f' {a:7.4f} {np.mean(c)/total:7.5f}\n')


laue_space_groups = 'P-1', 'P2/m', 'Pmmm', 'P4/m', 'P4/mmm', 'P-3', 'P-3m1',\
                    'P6/m', 'P6/mmm', 'Pm-3', 'Pm-3m'
laue_class_names = r"$\overline{1}$", r"$2/m$", r"$mmm$",\
                   r"$4/m$", r"$4/mmm$", r"$\overline{3}$", r"$\overline{3}m$",\
                   r"$6/m$", r"$6/mmm$", r"$m\overline{3}$", r"$m\overline{3}m$"


def potency_violin_plot(job_name='violin',
                        directory='~',
                        opening_angle=35,
                        precision=1000,
                        space_groups=laue_space_groups,
                        labels=laue_class_names,
                        resolution=None,
                        wavelength='MoKa'):
    """
    Calculate potency distribution for selected space groups and multiple sample
    orientations in a DAC, log it, and (re)draw appropriate violin plot.

    :param job_name: Base name for created files, defaults to 'violin'. If a
        .csv file with this name already exists, redraw instead of calculating.
    :type job_name: str
    :param directory: Target directory for saving or reading, defaults to '~'.
    :type directory: str
    :param opening_angle: Value of single opening angle as defined in
        :meth:`hikari.dataframes.HklFrame.dac`. Defaults to 35.
    :type opening_angle: float
    :param precision: Number of orientations to investigate, defaults to 1000.
    :type precision: int
    :param space_groups: List of space groups to investigate, strings or ints
        (see :class:`hikari.symmetry.Group`). Defaults to 11 "Laue" groups.
    :type space_groups: tuple[int] or tuple[str]
    :param labels: List of tex-style labels to be used for logging and plotting.
    :type labels: tuple[str]
    :param resolution: If given, additionally limit data resolution to given
        value. Please provide the resolution as a distance from the origin
        in reciprocal space (twice the resolution in reciprocal angstrom).
    :type resolution: float
    :param wavelength: Wavelength of radiation to be simulated.
    :type wavelength: float or str
    :return: None
    :rtype: None
    """

    csv_path = make_abspath(directory, job_name + '.csv')
    png_path = make_abspath(directory, job_name + '.png')
    log_path = make_abspath(directory, job_name + '.log')
    space_groups = [SG[sg] for sg in space_groups]

    def _is_tri_or_hexagonal(sg):
        return sg.system in {sg.System.hexagonal, sg.System.trigonal}

    def _make_ball(ga):
        hkl_frame = HklFrame()
        hkl_frame.edit_cell(a=20, b=20, c=20, al=90, be=90, ga=ga)
        hkl_frame.la = wavelength
        hkl_frame.fill(radius=resolution if resolution else hkl_frame.r_lim)
        return hkl_frame

    h = _make_ball(ga=120) if any(_is_tri_or_hexagonal(sg)
                                  for sg in space_groups) else None
    c = _make_ball(ga=90) if not all(_is_tri_or_hexagonal(sg)
                                     for sg in space_groups) else None
    vectors = np.vstack([np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1],
                                   [0.57735, 0.57735, 0.57735]]),
                         fibonacci_sphere(samples=precision-4, seed=1337)])

    try:
        cplt_frame = pd.read_csv(csv_path, index_col=0)
    except FileNotFoundError:
        cplt_dict = dict()
        log = open(log_path, 'w', buffering=1)
        for label, sg in zip(labels, space_groups):
            p = h.copy() if _is_tri_or_hexagonal(sg) else c.copy()
            p.find_equivalents(point_group=sg.reciprocate())
            p.extinct(space_group=sg)
            total_reflections = p.table['equiv'].nunique()
            log.write('space_group: ' + str(label) + '\n')
            log.write('total_reflections: ' + str(total_reflections) + '\n')
            log.write('max_r_in_reciprocal: ' + str(max(p.table['r'])) + '\n')
            reflections = p.dacs_count(opening_angle, vectors)
            for v, r in zip(vectors, reflections):
                log.write(str(v)+': '+str(r)+'\n')
            log.write('max_reflections: ' + str(max(reflections)) + '\n')
            log.write('min_reflections: ' + str(min(reflections)) + '\n')
            log.write('avg_reflections: ' + str(np.mean(reflections)) + '\n\n')
            cplt_dict[label] = [r / total_reflections for r in reflections]

        cplt_frame = pd.DataFrame.from_dict(cplt_dict)
        cplt_frame = cplt_frame.mul(100)
        cplt_frame.to_csv(csv_path)
        pd.set_option('display.expand_frame_repr', False)
        log.write('Summary:\n' + str(cplt_frame.describe().round(2)) + '\n')
        log.close()

    palette_number = {Group.System.triclinic: 0,
                      Group.System.monoclinic: 1,
                      Group.System.orthorhombic: 2,
                      Group.System.tetragonal: 3,
                      Group.System.trigonal: 4,
                      Group.System.hexagonal: 5,
                      Group.System.cubic: 6}
    palette = {label: sns.color_palette("deep")[palette_number[sg.system]]
               for label, sg in zip(labels, space_groups)}

    melt_frame = cplt_frame.melt(var_name='SG', value_name='Completeness')
    sns.set_style("whitegrid")
    pyplot.figure(figsize=(8, 6))
    sns.set(font_scale=1.5)
    ax = sns.violinplot(x=melt_frame['Completeness'],
                        y=melt_frame['SG'], scale='width', cut=0,
                        palette=palette, inner='quartile')
    for line in ax.lines[1::3]:
        line.set_linestyle('--')
        line.set_linewidth(1.25)
    ax.set_xlabel('')
    ax.set_ylabel('')
    for y, key in enumerate(cplt_frame.keys()):
        ax.axvline(x=cplt_frame[key].median(), color=palette[key],
                   ymin=0, ymax=21/22-y/11, linewidth=5, alpha=0, zorder=0.99)
    ax.set_yticklabels(labels)
    ax.set_xticklabels(['10', r'$\mathcal{P}$ = 20       ', '30', '40',
                        '50', '60', '70', '80', '90', '100%'])
    pyplot.xlim(19, 101)
    sns.despine(left=True, top=False)
    pyplot.savefig(png_path, dpi=600, format='png', bbox_inches=None)


def dac_potency_around_axis(a, b, c, al, be, ga,
                            space_group='P1',
                            opening_angle=35.0,
                            wavelength='MoKa',
                            vector=(1, 0, 0),
                            topple_angle=5):
    """
    For a given crystal, opening angle, and wavelength calculate average potency
    obtained by toppling the crystal by "topple_angle"° from the "vector" axis.

    :param a: Unit cell parameter *a* in Angstrom.
    :type a: float
    :param b: Unit cell parameter *b* in Angstrom.
    :type b: float
    :param c: Unit cell parameter *c* in Angstrom.
    :type c: float
    :param al: Unit cell parameter *alpha* in degrees.
    :type al: float
    :param be: Unit cell parameter *beta* in degrees.
    :type be: float
    :param ga: Unit cell parameter *gamma* in degrees.
    :type ga: float
    :param space_group: Short Hermann-Mauguin name or index of space group.
        For details see :py:mod:`hikari.symmetry.space_groups`.
    :type space_group: str or int
    :param wavelength: Wavelength of radiation utilised in experiment.
    :type wavelength: float or str
    :param opening_angle: Value of single opening angle as defined in
        :meth:`hikari.dataframes.HklFrame.dac`.
    :type opening_angle: float
    :param vector: Direction from which a theoretical crystal will be toppled.
    :type vector: tuple
    :param topple_angle: Angle by which theoretical crystal will be toppled.
    :type topple_angle: float
    :return: None
    """

    sg = SG[space_group]
    pg = sg.reciprocate()  # .lauefy()  # uncomment if hkl and -h-k-l are equiv.

    p = HklFrame()
    p.edit_cell(a=a, b=b, c=c, al=al, be=be, ga=ga)
    p.la = wavelength
    p.fill(radius=p.r_lim)
    p.place()
    p.extinct(space_group=sg)
    p.find_equivalents(point_group=pg)

    # generate perpendicular vector for rotation
    v = np.array(vector) / lin.norm(np.array(vector))
    x = np.array((1, 0, 0))

    def are_parallel(v1, v2):
        return np.allclose(v1 / lin.norm(v1), v2 / lin.norm(v2))
    temp = x if not(are_parallel(v, x)) else np.array((0, 1, 0))
    perp = np.cross(v, temp)

    def rotate(_v, _k, angle):
        return _v @ rotation_around(_k, by=np.deg2rad(angle))

    # generate 360 toppled vectors
    toppled_vector = rotate(v, perp, topple_angle)
    toppled_vectors = [rotate(toppled_vector, v, i) for i in range(360)]

    # calculate the potency for toppleds
    rads = [0.6]
    cplt = [0] * len(rads)
    p.trim(max(rads))
    for t in toppled_vectors:
        q = p.copy()
        q.dac_trim(opening_angle=opening_angle, vector=t)
        for cplt_bin, rad in enumerate(rads, start=1):
            q.trim(rad)
            cplt[-cplt_bin] += q.table['equiv'].nunique()

    # divide lists of potency by total and return from sums to individuals
    full = [0] * len(rads)
    for cplt_bin, rad in enumerate(rads, start=1):
        p.trim(rad)
        full[-cplt_bin] = p.table['equiv'].nunique()
    print(cplt)
    print(full)
    cplt = [c2 - c1 for c1, c2 in zip([0] + cplt[:-1], cplt)]
    full = [f2 - f1 for f1, f2 in zip([0] + full[:-1], full)]
    cplt = [c / (360 * f) for c, f in zip(cplt, full)]

    print('Resolution limits in distance to reflection, twice sin(θ/λ):')
    print(np.array(list(reversed(rads))))
    print(f'Average shell potency with {topple_angle} degree topple')
    print(np.array(cplt))


if __name__ == '__main__':
    potency_map(a=10, b=10, c=10, al=90, be=90, ga=120, space_group='P21/c',
                path='~/_/potency.hkl', output_quality=5, wavelength='MoKa')



"""
This file contains constants and functions containing operating system and path
information / methods used in the package.
"""

import pathlib


def make_abspath(*path_elements):
    """
    Return a string with absolute path to a specified file. Accepts zero or more
    strings, which are joined according to system-specific syntax. Relative
    paths are intepreted as having root at current working directory (cwd).
    Links and symbols such as '~', '..' or '/' are expanded and interpreted.
    As such, function calls with no argument and '~' as argument will yield
    string-paths to current working directory and home directory, respectively.

    :Example:

    >>> make_abspath()
    PosixPath('/home/username/hikari/hikari/utility')
    >>> make_abspath('~')
    PosixPath('/home/username')
    >>> make_abspath('~', 'cocoa/is', 'not/../hot.txt')
    PosixPath('/home/username/cocoa/is/hot.txt')

    Running the script on Windows machine will yield an appropriate Windows
    Path. In particular, 'cocoa/is' will be intepreted as single directory.

    :param path_elements: path element(s) to join and intepret, defaults to cwd.
    :type path_elements: str
    :return: string containing resolved absolute path to the specified location
    :rtype: str
    """
    return str(pathlib.Path().joinpath(*path_elements).expanduser().resolve())



from enum import Enum
from collections import defaultdict
import numpy as np
from hikari.dataframes import BaseFrame
from hikari.resources import Xray_atomic_form_factors
from hikari.utility import split_atom_label, make_abspath

res_instructions = defaultdict(set)


class ResInstructionType(Enum):
    UNIQUE_INT = 1
    UNIQUE_LIST = 2
    UNIQUE_MULTILINE = 3
    REPEATING_SINGLET = -1


class ResFrame(BaseFrame):
    def __init__(self):
        super().__init__()
        self.data = {}

    def atomic_form_factor(self, atom, hkl):
        """
        Calculate X-ray atomic form factors for a single atom and a hkl array

        :param atom: Atom/ion name/identifier interpreted by form factor table
        :type atom: str
        :param hkl: A 2D array listing all hkls to consider
        :type hkl: np.array
        :return: A 1D array listing atomic form factors for desired hkls
        :rtype: np.array
        """
        r_star = self.A_r @ hkl.T
        s = Xray_atomic_form_factors.loc[atom]
        sintl2 = (r_star * r_star).sum(axis=0) / 4
        return s['a1'] * np.exp(-s['b1'] * sintl2) + \
               s['a2'] * np.exp(-s['b2'] * sintl2) + \
               s['a3'] * np.exp(-s['b3'] * sintl2) + \
               s['a4'] * np.exp(-s['b4'] * sintl2) + s['c']

    def temperature_factor(self, hkl, u):
        """
        Calculate temperature factor for single u matrix and a hkl array

        :param hkl: A 2D array listing all hkls to consider
        :type hkl: np.array
        :param u: A classical anisotropic displacement parameters matrix
        :type u: np.array
        :return: A 1D array listing temperature factors for desired hkls
        :rtype: np.array
        """
        n_matrix = np.diag([self.a_r, self.b_r, self.c_r])
        hkl_n_u_n_hkl = (hkl @ n_matrix * hkl @ n_matrix @ u.T).sum(axis=-1)
        return np.exp(-2 * np.pi ** 2 * hkl_n_u_n_hkl)

    def form_factor(self, hkl, space_group):
        """
        Calculate form factors based on current structure, hkls, and space group

        :param hkl: A 2D array listing all hkls to consider
        :type hkl: np.array
        :param space_group: Space group describing the internal crystal symmetry
        :type space_group: hikari.symmetry.Group
        :return: A 1D array listing total form factors for desired hkls
        :rtype: np.array
        """
        f = np.zeros_like(hkl.sum(axis=-1), dtype='complex128')
        for k, v in self.data['ATOM'].items():
            atom = split_atom_label(k)[0].title()
            xyz = np.fromiter(map(float, v[1:4]), dtype=np.float)
            occupation = float(v[4]) % 10
            uij_strings = v[5:] if len(v[5:]) == 6 else [v[5]] * 3 + ['0'] * 3
            u11, u22, u33, u23, u13, u12 = map(float, uij_strings)
            u = np.array([[u11, u12, u13], [u12, u22, u23], [u13, u23, u33]])
            for o in space_group.operations:
                xyz_t = (o.tf @ np.array(xyz)).T + o.tl
                u_t = o.tf @ u @ (o**-1).tf
                fa = occupation * self.atomic_form_factor(atom, hkl) * \
                    self.temperature_factor(hkl, u_t)
                f += fa * np.exp(2j * np.pi * xyz_t @ hkl.T)
        return f
    # TODO the factors seem to be evaluated incorrectly, check in the future

    def read(self, path):
        """
        Read data from specified ins/res file and return an dict

        :param path: Relative or absolute path to the res file to be read
        :type path: str
        :return: None
        :rtype: None
        """

        # SPECIFY DEFAULT VALUES OF NECESSARY KEYS
        self.data['REM'] = ['These comments were found by resins:']
        self.data['ATOM'] = dict()
        self.data['PEAK'] = dict()
        self.data['TITL'] = list()
        self.data['SYMM'] = list()

        # SPECIFY SUPPORTED KEYS AND THEIR TYPES
        key_types = {
            'TITL': 'multiline',
            'CELL': 'listing',
            'ZERR': 'listing',
            'LATT': 'listing',
            'SYMM': 'multiline',
            'SFAC': 'listing',
            'UNIT': 'listing',
            'L.S.': 'listing',
            'PLAN': 'listing',
            'SIZE': 'listing',
            'MORE': 'listing',
            'BOND': 'listing',
            'CONF': 'listing',
            'FMAP': 'listing',
            'ACTA': 'listing',
            'WGHT': 'listing',
            'FVAR': 'listing',
            'HKLF': 'listing',
            'REM': 'special',
            'END': 'special',
        }

        # READ THE FILE AND JOIN LINES SEPARATED BY '=' SIGN
        with open(make_abspath(path), 'r') as res_file:
            lines = res_file.read().replace('=\n', '').split('\n')
            lines = [line.strip() for line in lines if line.strip()]

        class ResIoStage(Enum):
            TITLE = 0
            PREAMBLE = 1
            ATOMS = 2
            APPENDIX = 3
            END = 4

        # READ THE FILE AND GATHER ALL COMMENTS, ATOMS AND PEAKS
        reading_stage = ResIoStage.PREAMBLE
        lines_to_delete = []
        for index, line in enumerate(lines):
            key = line.split(maxsplit=1)[0]
            values = list(line[len(key):].strip().split())
            if key.upper() == 'FVAR':
                reading_stage = ResIoStage.ATOMS
            elif key.upper() == 'HKLF':
                reading_stage = ResIoStage.APPENDIX
            elif key.upper() == 'END':
                reading_stage = ResIoStage.END
            elif key.upper() == 'REM':
                self.data['REM'].append(line)
                lines_to_delete.append(index)
            elif all((key.upper() not in key_types.keys(),
                      reading_stage is ResIoStage.END,
                      key[0] == 'Q', key[1].isdigit())):
                self.data['PEAK'][key] = values
                lines_to_delete.append(index)
            elif all((key.upper() not in key_types.keys(),
                      reading_stage is ResIoStage.ATOMS,
                      key[0].isalpha())):
                self.data['ATOM'][key] = values
                lines_to_delete.append(index)
            elif key.upper() not in key_types.keys():
                self.data['REM'].append('Line not interpreted: '+line)
                lines_to_delete.append(index)
        for index in reversed(lines_to_delete):
            del lines[index]

        # FOR EACH LINE SPLIT LINE TO KEY AND VALUE AND APPEND SELF
        for line in lines:
            key = line.split(maxsplit=1)[0].upper()
            value = line[len(key):].strip()
            if key_types[key] == 'listing':
                self.data[key] = list(value.split())
            elif key_types[key] == 'string':
                self.data[key] = value
            elif key_types[key] == 'multiline':
                self.data[key].append(value)
            elif key == 'END':
                break

        # BRING REM, ATOM AND PEAK TO THE END AND RETURN DICTIONARY
        self.data.move_to_end('REM')
        self.data.move_to_end('ATOM')
        self.data.move_to_end('PEAK')



import numpy as np
from hikari.utility import angle2rad, cfloat, det3x3
from hikari.utility import str2array as s2a


class BaseFrame:
    """
    This class stores and manipulates basic information present
    in majority of crystallographic information files such as unit cell
    parameters stored in scalars and vectors.

    BaseFrame utilises the following notation for stored attributes:

    - The name begins from a unit cell property we are interested in:

     - "a", "b", "c" describe unit cell lengths/vectors *a*, *b*, *c*,
     - "al", "be", "ga" describe unit cell angles *alpha*, *beta*, *gamma*,
     - "v" describes unit cell volume,
     - "x", "y", "z" describe directions - normalised unit cell vectors.
     - "A", "G" describe stacked vector and metric matrix, respectively.

    - The unit cell parameter symbol is then followed by an underscore "_".

    - The name ends with a single letter denoting type of space and variable:

     - "d" (from Direct) denotes direct space scalars/matrices,
     - "r" (from Reciprocal) denotes reciprocal space scalars/matrices,
     - "v" (from Vector) denotes direct space vectors,
     - 'w" (similar to "v") denotes reciprocal space vectors.

    The values can be accessed by referencing a given attribute in the object,
    for example :class:`BaseFrame`. :attr:`a_d` stores information about
    the lattice constant *a* in direct space as a floating point,
    but :class:`BaseFrame`. :attr:`a_v` is a direct space vector.
    Available attributes have been once again presented in a table below:

    +----------+------------+------------------+------------------+------------+
    |          | Available  | in direct        | in reciprocal    |Unit (^-1 in|
    |          | constants  | space            | space            |reciprocal) |
    +==========+============+==================+==================+============+
    | Scalars  | a, b, c    | a_d, b_d, c_d    | a_r, b_r, c_r    | Angstrom   |
    |          +------------+------------------+------------------+------------+
    |          | al, be, ga | al_d, be_d, ga_d | al_r, be_r, ga_r | Radian     |
    |          +------------+------------------+------------------+------------+
    |          | v          | v_d              | v_r              | Angstrom^3 |
    +----------+------------+------------------+------------------+------------+
    | Vectors  | a, b, c    | a_v, b_v, c_v    | a_w, b_w, c_w    | Angstrom   |
    |          +------------+------------------+------------------+------------+
    |          | x, y, z    | x_v, y_v, z_v    | x_w, y_w, z_w    | Angstrom   |
    +----------+------------+------------------+------------------+------------+
    | Matrices | A          | A_d              | A_r              | Angstrom^2 |
    |          +------------+------------------+------------------+------------+
    |          | G          | G_d              | G_r              | Angstrom^2 |
    +----------+------------+------------------+------------------+------------+
    """

    IMPORTED_FROM_CIF = {
        'a': ['_cell_length_a', cfloat, 1.0],
        'b': ['_cell_length_b', cfloat, 1.0],
        'c': ['_cell_length_c', cfloat, 1.0],
        'al': ['_cell_angle_alpha', cfloat, 90],
        'be': ['_cell_angle_beta', cfloat, 90],
        'ga': ['_cell_angle_gamma', cfloat, 90],
        'ub11': ['_diffrn_orient_matrix_UB_11', float, 1.0],
        'ub12': ['_diffrn_orient_matrix_UB_12', float, 0.0],
        'ub13': ['_diffrn_orient_matrix_UB_13', float, 0.0],
        'ub21': ['_diffrn_orient_matrix_UB_21', float, 0.0],
        'ub22': ['_diffrn_orient_matrix_UB_22', float, 1.0],
        'ub23': ['_diffrn_orient_matrix_UB_23', float, 0.0],
        'ub31': ['_diffrn_orient_matrix_UB_31', float, 0.0],
        'ub32': ['_diffrn_orient_matrix_UB_32', float, 0.0],
        'ub33': ['_diffrn_orient_matrix_UB_33', float, 1.0]}

    def __init__(self):
        self._a_d = self._b_d = self._c_d = 1.0
        self._a_r = self._b_r = self._c_r = 1.0
        self._al_d = self._be_d = self._ga_d = np.pi / 2
        self._al_r = self._be_r = self._ga_r = np.pi / 2
        self._a_v, self._b_v, self._c_v = np.eye(3)
        self._a_w, self._b_w, self._c_w = np.eye(3)
        self._refresh_cell()
        self.orientation = np.array(((1.0, 0, 0), (0, 1.0, 0), (0, 0, 1.0)))
        """3x3 matrix describing orientation of crystal during experiment."""

    def edit_cell(self, **parameters):
        """
        Edit direct space unit cell using a dictionary with the following keys:

        - "a" - for unit cell parameter *a* given in Angstrom,
        - "b" - for unit cell parameter *b* given in Angstrom,
        - "c" - for unit cell parameter *c* given in Angstrom,
        - "al" - for unit cell parameter *alpha* given in degrees or radians,
        - "be" - for unit cell parameter *beta* given in degrees or radians,
        - "ga" - for unit cell parameter *gamma* given in degrees or radians.

        This method is equivalent to manually setting all six unit cell
        parameters in direct space, :attr:`a_d`, :attr:`b_d`, :attr:`c_d`,
        :attr:`al_d`, :attr:`be_d`, :attr:`ga_d`, and then running a private
        method :meth:`_refresh_cell` to update other values.

        Please mind that the while the "a", "b" and "c" are always given in
        Angstrom, the angles might be given either in degrees or in radians.
        For details see function :func:`hikari.utility.math_tools.angle2rad`.

        It is not required for all previously stated keys to be present
        at each method call. If a key has not been given, previously provided
        and stored value is being used. If no value has been given,
        the default length values of 1.0 for *a*, *b*, *c* and default angle
        values of *pi/2* for *al*, *be*, *ga* are used instead.

        :param parameters: Values of unit cell parameters to be changed
        :type parameters: float
        """
        for key, value in parameters.items():
            if key not in {'a', 'b', 'c', 'al', 'be', 'ga'}:
                raise KeyError(f'Unknown unit cell parameter: {key}')
            elif key in {'a', 'b', 'c'}:
                setattr(self, f'_{key}_d', value)
            elif key in {'al', 'be', 'ga'}:
                setattr(self, f'_{key}_d', angle2rad(value))
        self._refresh_cell()

    def fill_from_cif_block(self, block, fragile=False):
        """
        Import all data specified in :attr:`~.BaseFrame.IMPORTED_FROM_CIF`
        such as unit cell parameters and orientation matrix from provided
        instance of :class:`hikari.dataframes.cif.CifBlock` called `block`.
        Unless `fragile` is `True`, use defaults instead of rising KeyError.

        :param block: CifBlock containing imported information.
        :type block: hikari.dataframes.CifBlock
        :param fragile: If True, raise Error when any imported info is missing
        :type fragile: bool
        """
        imp = {}
        for k, v in self.IMPORTED_FROM_CIF.items():
            if fragile:
                imp[k] = block.get_as_type(v[0], v[1], None)
            else:
                imp[k] = block.get_as_type(v[0], v[1], v[2])
        if None in imp.values():
            miss = [k for k, v in imp.items() if v is None]
            raise KeyError(f'Missing values when filling from CifBlock: {miss}')
        cell_par_names = {'a', 'b', 'c', 'al', 'be', 'ga'}
        new_parameters = {k: v for k, v in imp.items() if k in cell_par_names}
        self.edit_cell(**new_parameters)
        orientation_value_names = ('ub11', 'ub12', 'ub13',
                                   'ub21', 'ub22', 'ub23',
                                   'ub31', 'ub32', 'ub33')
        orientation_values = [imp[n] for n in orientation_value_names]
        self.orientation = np.array(orientation_values).reshape(3, 3)

    def _refresh_cell(self):
        """
        Recalculate all vectors and scalars other than :attr:`a_d`,
        :attr:`b_d`, :attr:`c_d`, :attr:`al_d`, :attr:`be_d`, :attr:`ga_d`
        based on the currently stored values of the aforementioned six.
        """
        a, b, c = self.a_d, self.b_d, self.c_d
        sa, sb, sg = np.sin(self.al_d), np.sin(self.be_d), np.sin(self.ga_d)
        ca, cb, cg = np.cos(self.al_d), np.cos(self.be_d), np.cos(self.ga_d)
        v = a * b * c * np.sqrt(1 - ca**2 - cb**2 - cg**2 + 2 * ca * cb * cg)

        self._a_v = np.array([a, 0, 0])
        self._b_v = np.array([b * cg, b * sg, 0])
        self._c_v = np.array([c * cb, c * (ca - cb * cg) / sg, v / (a * b * sg)])

        self._a_r = b * c * sa / v
        self._b_r = c * a * sb / v
        self._c_r = a * b * sg / v
        self._al_r = np.arccos((cb * cg - ca) / (sb * sg))
        self._be_r = np.arccos((cg * ca - cb) / (sg * sa))
        self._ga_r = np.arccos((ca * cb - cg) / (sa * sb))

        self._a_w = np.cross(self.b_v, self.c_v) / v
        self._b_w = np.cross(self.c_v, self.a_v) / v
        self._c_w = np.cross(self.a_v, self.b_v) / v

    @property
    def a_d(self):
        """
        :return: Length of unit cell vector **a** in direct space.
        :rtype: float
        """
        return self._a_d

    @property
    def b_d(self):
        """
        :return: Length of unit cell vector **b** in direct space.
        :rtype: float
        """
        return self._b_d

    @property
    def c_d(self):
        """
        :return: Length of unit cell vector **c** in direct space.
        :rtype: float
        """
        return self._c_d

    @property
    def al_d(self):
        """
        :return: Angle between vectors **b** and **c** in degrees.
        :rtype: float
        """
        return self._al_d

    @property
    def be_d(self):
        """
        :return: Angle between vectors **c** and **a** in degrees.
        :rtype: float
        """
        return self._be_d

    @property
    def ga_d(self):
        """
        :return: Angle between vectors **a** and **b** in degrees.
        :rtype: float
        """
        return self._ga_d

    @property
    def v_d(self):
        """
        :return: Unit cell volume in direct space.
        :rtype: float
        """
        return det3x3(self.A_d)

    @property
    def a_v(self):
        """
        :return: Unit cell vector **a** in direct space.
        :rtype: numpy.array
        """
        return self._a_v

    @property
    def b_v(self):
        """
        :return: Unit cell vector **b** in direct space.
        :rtype: numpy.array
        """
        return self._b_v

    @property
    def c_v(self):
        """
        :return: Unit cell vector **c** in direct space.
        :rtype: numpy.array
        """
        return self._c_v

    @property
    def A_d(self):
        """
        :return: Basis matrix A with vertically stacked direct space vectors.
        :rtype: np.array
        """
        return np.vstack([self._a_v, self._b_v, self._c_v])

    @property
    def G_d(self):
        """
        :return: Direct space metric matrix [ai . aj]ij.
        :rtype: np.array
        """
        return self.A_d @ self.A_d.T

    @property
    def a_r(self):
        r"""
        :return: Length of unit cell vector **a\*** in reciprocal space.
        :rtype: float
        """
        return self._a_r

    @property
    def b_r(self):
        r"""
        :return: Length of unit cell vector **b\*** in reciprocal space.
        :rtype: float
        """
        return self._b_r

    @property
    def c_r(self):
        r"""
        :return: Length of unit cell vector **c\*** in reciprocal space.
        :rtype: float
        """
        return self._c_r

    @property
    def al_r(self):
        r"""
        :return: Angle between vectors **b\*** and **c\*** in degrees.
        :rtype: float
        """
        return self._al_r

    @property
    def be_r(self):
        r"""
        :return: Angle between vectors **c\*** and **a\*** in degrees.
        :rtype: float
        """
        return self._be_r

    @property
    def ga_r(self):
        r"""
        :return: Angle between vectors **a\*** and **b\*** in degrees.
        :rtype: float
        """
        return self._ga_r

    @property
    def v_r(self):
        """
        :return: Unit cell volume in reciprocal space.
        :rtype: float
        """
        return det3x3(self.A_r)

    @property
    def a_w(self):
        r"""
        :return: Unit cell vector **a\*** in reciprocal space.
        :rtype: numpy.array
        """
        return self._a_w

    @property
    def b_w(self):
        r"""
        :return: Unit cell vector **b\*** in reciprocal space.
        :rtype: numpy.array
        """
        return self._b_w

    @property
    def c_w(self):
        r"""
        :return: Unit cell vector **c\*** in reciprocal space.
        :rtype: numpy.array
        """
        return self._c_w

    @property
    def A_r(self):
        r"""
        :return: Basis matrix A\* with vertically stacked reciprocal space vectors.
        :rtype: np.array
        """
        return np.vstack([self._a_w, self._b_w, self._c_w])

    @property
    def G_r(self):
        r"""
        :return: Reciprocal space metric matrix [ai\* . aj\*]ij.
        :rtype: np.array
        """
        return self.A_r @ self.A_r.T

    SELLING_S6_TRANSFORMATIONS = [
        s2a('-100000/110000/100010/-100100/101000/100001'),
        s2a('110000/0-10000/010100/011000/0-10010/010001'),
        s2a('101000/001100/00-1000/011000/001010/00-1001'),
        s2a('100-100/001100/010100/000-100/000110/000101'),
        s2a('001010/0100-10/100010/000110/0000-10/000011'),
        s2a('010001/100001/00100-1/000101/000011/00000-1')]

    SELLING_E3_TRANSFORMATIONS = [
        s2a('110/0-10/001'), s2a('-100/110/001'), s2a('-100/010/101'),
        s2a('-100/110/101'), s2a('110/0-10/011'), s2a('101/011/00-1')]



"""
Module containing symmetry operations and point groups
to be used by other objects.
"""

from .operations import Operation, BoundedOperation, PointOperation
from .group import Group
from .hall import HallSymbol
from .catalog import GroupCatalog, PG, SG



gnuplot_map_palette = dict()
gnuplot_map_palette[''] = 'set palette model HSV defined (0 0.667 1 0,' + \
                          ' 20 0.667 1 1, 80 0.0 1 1, 100 0.0 0 1)'
gnuplot_map_palette['h'] = 'set palette model RGB defined (0 1 1 1, 1 1 0 0)'
gnuplot_map_palette['x'] = 'set palette model RGB defined (0 0 0 0, 1 1 0 0)'
gnuplot_map_palette['k'] = 'set palette model RGB defined (0 1 1 1, 1 0 1 0)'
gnuplot_map_palette['y'] = 'set palette model RGB defined (0 0 0 0, 1 0 1 0)'
gnuplot_map_palette['l'] = 'set palette model RGB defined (0 1 1 1, 1 0 0 1)'
gnuplot_map_palette['z'] = 'set palette model RGB defined (0 0 0 0, 1 0 0 1)'
gnuplot_map_palette['hk'] = gnuplot_map_palette['l']
gnuplot_map_palette['xy'] = gnuplot_map_palette['z']
gnuplot_map_palette['kl'] = gnuplot_map_palette['h']
gnuplot_map_palette['yz'] = gnuplot_map_palette['x']
gnuplot_map_palette['hl'] = gnuplot_map_palette['k']
gnuplot_map_palette['xz'] = gnuplot_map_palette['y']

mpl_map_palette = dict()
mpl_map_palette[''] = [(0.0, 0.0, 0.0),  # Black
                       (0.0, 0.0, 0.5),  # Indigo
                       (0.0, 0.0, 1.0),  # Blue
                       (0.0, 0.5, 1.0),  # Aqua
                       (0.0, 1.0, 1.0),  # Turquoise
                       (0.5, 1.0, 0.5),  # Lime
                       (1.0, 1.0, 0.0),  # Yellow
                       (1.0, 0.5, 0.0),  # Orange
                       (1.0, 0.0, 0.0),  # Red
                       (1.0, 0.5, 0.5),  # Light red
                       (1.0, 1.0, 1.0)]  # White
mpl_map_palette['h'] = [(1.0, 1.0, 1.0), (1.0, 0.0, 0.0)]
mpl_map_palette['x'] = [(0.0, 0.0, 0.0), (1.0, 0.0, 0.0)]
mpl_map_palette['k'] = [(1.0, 1.0, 1.0), (0.0, 1.0, 0.0)]
mpl_map_palette['y'] = [(0.0, 0.0, 0.0), (0.0, 1.0, 0.0)]
mpl_map_palette['l'] = [(1.0, 1.0, 1.0), (0.0, 0.0, 1.0)]
mpl_map_palette['z'] = [(0.0, 0.0, 0.0), (0.0, 0.0, 1.0)]
mpl_map_palette['hk'] = mpl_map_palette['l']
mpl_map_palette['xy'] = mpl_map_palette['z']
mpl_map_palette['kl'] = mpl_map_palette['h']
mpl_map_palette['yz'] = mpl_map_palette['x']
mpl_map_palette['hl'] = mpl_map_palette['k']
mpl_map_palette['xz'] = mpl_map_palette['y']



"""
This file contains class definition and necessary tools for constructing
and evaluating all symmetry groups.
"""
from itertools import product as itertools_product
from enum import Enum
from typing import Union

import numpy as np
from numpy.random.mtrand import Sequence

from hikari.symmetry.operations import BoundedOperation
from hikari.symmetry.hall import HallSymbol
from hikari.utility.list_tools import find_best


class Group:
    """
    Base immutable class containing information about symmetry groups.
    It stores information for point and space groups and, among others,
    allows for iteration over its `hikari.symmetry.BoundedOperation` elements.
    """

    class System(Enum):
        """Enumerator class with information about associated crystal system"""
        triclinic = 0
        monoclinic = 1
        orthorhombic = 2
        trigonal = 3
        tetragonal = 4
        cubic = 5
        hexagonal = 6

        @property
        def directions(self) -> list[np.ndarray]:
            _a = np.array((1, 0, 0))
            _b = np.array((0, 1, 0))
            _c = np.array((0, 0, 1))
            _ab = np.array((1 / np.sqrt(2), 1 / np.sqrt(2), 0))
            _abc = np.array((1 / np.sqrt(3), 1 / np.sqrt(3), 1 / np.sqrt(3)))
            return [[], [_b, ], [_a, _b, _c], [_c, _a, _ab],
                    [_c, _a, _ab], [_c, _abc, _ab], [_c, _a, _ab]][self.value]

    BRAVAIS_PRIORITY_RULES = 'A+B+C=F>R>I>C>B>A>H>P'
    AXIS_PRIORITY_RULES = '6>61>62>63>64>65>-6>4>41>42>43>-4>-3>3>31>32>2>21'
    PLANE_PRIORITY_RULES = 'm>a+b=e>a+c=e>b+c=e>a>b>c>n>d'

    def __init__(self, *generators: BoundedOperation) -> None:
        """
        :param generators: List of operations necessary to construct whole group
        """

        generator_list = []
        for gen in generators:
            if gen not in generator_list:
                generator_list.append(gen)

        def _find_new_product(ops: list[BoundedOperation]) -> list[BoundedOperation]:
            if len(ops) > 200:
                raise ValueError('Generated group order exceeds size of 200')
            new = list({o1 * o2 for o1, o2 in itertools_product(ops, ops)})
            new = list(set(ops).union(new))
            return _find_new_product(new) if len(new) > len(ops) else ops

        self.__generators = list(generator_list)
        self.__operations = list(_find_new_product(generator_list))
        self.name = self.auto_generated_name
        self.number = 0

    @classmethod
    def from_generators_operations(
            cls,
            generators: list[BoundedOperation],
            operations: list[BoundedOperation],
    ) -> 'Group':
        """
        Generate group using already complete list of generators and operators.
        Does not check if `operations` are correct or complete for efficiency!
        :param generators: A complete list of group generators
        :param operations: A complete list of group operations
        :return: Symmetry group with given generators and operators.
        """
        new_group = cls()
        new_group.__generators = generators
        new_group.__operations = operations
        return new_group

    @classmethod
    def from_hall_symbol(cls, hall_symbol: Union[str, HallSymbol]) -> 'Group':
        if isinstance(hall_symbol, str):
            hall_symbol = HallSymbol(hall_symbol)
        return cls(*hall_symbol.generators)

    def __eq__(self, other: 'Group') -> bool:
        return all([o in self.operations for o in other.operations])\
               and all([o in other.operations for o in self.operations])

    def __lt__(self, other: 'Group') -> bool:
        return len(self.operations) < len(other.operations) and \
            all([o in other.operations for o in self.operations])

    def __gt__(self, other: 'Group') -> bool:
        return other.__lt__(self)

    def __le__(self, other: 'Group') -> bool:
        return self.__eq__(other) or self.__lt__(other)

    def __ge__(self, other: 'Group') -> bool:
        return self.__eq__(other) or other.__lt__(self)

    def __repr__(self) -> str:
        return 'Group('+',\n      '.join([repr(g) for g in self.generators])+')'

    def __str__(self) -> str:
        return f'{self.name} (#{abs(self.number)}{"*" if self.number<0 else""})'

    def __hash__(self) -> int:
        return sum(hash(o) for o in self.operations)

    @property
    def auto_generated_name(self) -> str:
        """Name of the group generated automatically. Use only as approx."""
        # TODO: enantiomorphs like P41 / P43 not recognised
        # TODO: 'e' found always whenever 'a' and 'b' present
        # TODO: some mistakes occur in trigonal crystal system (see SG149+)
        name = self.centering_symbol
        for d in self.system.directions:
            ops = [o.name.partition(':')[0] for o in self.operations
                   if o.orientation is not None and
                   np.isclose(abs(np.dot(np.abs(o.orientation), np.abs(d))), 1)]
            best_axis = find_best(ops, self.AXIS_PRIORITY_RULES)
            best_plane = find_best(ops, self.PLANE_PRIORITY_RULES)
            sep = '/' if len(best_axis) > 0 and len(best_plane) > 0 else ''
            name += '_' + best_axis + sep + best_plane
        return name.strip()

    @property
    def centering_symbol(self) -> str:
        tl = ([o.name for o in self.operations if o.typ is o.Type.translation])
        tl.append('H' if self.system is self.System.trigonal else 'P')
        return find_best(tl, self.BRAVAIS_PRIORITY_RULES)

    @property
    def generators(self) -> list[BoundedOperation]:
        return self.__generators

    @property
    def operations(self) -> list[BoundedOperation]:
        return self.__operations

    @property
    def order(self) -> int:
        return len(self.__operations)

    @property
    def is_centrosymmetric(self) -> bool:
        """True if group has centre of symmetry; False otherwise."""
        return any(np.isclose(op.trace, -3) for op in self.operations)

    @property
    def is_enantiogenic(self) -> bool:
        """True if determinant of any operation in group is negative."""
        return any(op.det < 0 for op in self.operations)

    @property
    def is_sohncke(self) -> bool:
        """True if determinant of all operations in group are positive."""
        return all(op.det > 0 for op in self.operations)

    @property
    def is_achiral(self):      # TODO See dictionary.iucr.org/Chiral_space_group
        return NotImplemented  # TODO and dx.doi.org/10.1524/zkri.2006.221.1.1

    @property
    def is_chiral(self):       # TODO See dictionary.iucr.org/Chiral_space_group
        return NotImplemented  # TODO and dx.doi.org/10.1524/zkri.2006.221.1.1

    @property
    def is_symmorphic(self) -> bool:
        zero_vector = np.array([0, 0, 0])
        trans = [o for o in self.operations if o.typ is o.Type.translation]
        zero_tl = [o for o in self.operations if np.allclose(o.tl, zero_vector)]
        return self.order == len(zero_tl) * (len(trans) + 1)

    @property
    def is_polar(self) -> bool:
        if any(op.typ in {op.Type.rotoinversion, op.Type.inversion}
               for op in self.operations):
            return False
        axes_orientation = [op.orientation for op in self.operations
                            if op.det > 0 and op.orientation is not None]
        for or1, or2 in itertools_product(axes_orientation, axes_orientation):
            if not np.isclose(abs(np.dot(or1, or2)), 1):
                return False
        return True

    @property
    def system(self) -> System:
        """Predicted crystal system associated with this group"""
        folds = [op.fold for op in self.operations]
        orients = [op.orientation for op in self.operations]

        def _is_many(_orients):
            return any([abs(np.dot(_orients[0], o)) < 0.99 for o in _orients[1:]])

        if 6 in folds:
            return self.System.hexagonal
        elif 3 in folds:
            orients_of_3 = [o for f, o in zip(folds, orients) if f == 3]
            return self.System.cubic if _is_many(orients_of_3) \
                else self.System.trigonal
        elif 4 in folds:
            return self.System.tetragonal
        elif 2 in folds:
            orients_of_2 = [o for f, o in zip(folds, orients) if f == 2]
            return self.System.orthorhombic if _is_many(orients_of_2) \
                else self.System.monoclinic
        else:
            return self.System.triclinic

    def lauefy(self) -> 'Group':
        """
        :return: New PointGroup with centre of symmetry added to generators.
        :rtype: Group
        """
        inv = BoundedOperation.from_code('-x, -y, -z')
        return Group(*self.operations, inv)

    def reciprocate(self) -> 'Group':
        new_generators = [op.reciprocal for op in self.generators]
        return Group(*new_generators)

    def transform(self, m) -> 'Group':
        r"""
        Transform the group using 4x4 matrix. For reference, see `bilbao
        resources <https://www.cryst.ehu.es/cryst/trmatrix.html>`_ or `IUCr
        pamphlet no. 22 <https://www.iucr.org/education/pamphlets/22>`_.

        :example:

        >>> import numpy
        >>> from hikari.symmetry import SG
        >>> matrix = numpy.array([(1,0,1,0),(0,1,0,0),(-1,0,0,0),(0,0,0,1)])
        >>> SG['P21/c'].transform(matrix).auto_generated_name
        P 21/n

        :param m: A 4x4 array containing information about new base and origin.
        :type m: np.ndarray
        :return: Group with new, transformed basis and origin.
        :rtype: Group
        """
        transformed_group = Group.from_generators_operations(
            generators=[BoundedOperation.from_matrix(np.linalg.inv(m) @ g.matrix @ m)
                        for g in self.generators],
            operations=[BoundedOperation.from_matrix(np.linalg.inv(m) @ o.matrix @ m)
                        for o in self.operations])
        transformed_group.name = self.name+' @ '+repr(m)[6:-1].replace(' ', '')
        transformed_group.number = -abs(self.number)
        return transformed_group



import copy
import random
from typing import Union, Iterable

import numpy as np
import numpy.linalg as lin
import pandas as pd

import hikari
from hikari.dataframes import BaseFrame
from hikari.symmetry import PG, SG, Group
from hikari.resources import hkl_formats, hkl_aliases, hkl_mercury_style, \
    characteristic_radiation
from hikari.utility import cubespace, chemical_elements, make_abspath, \
    rescale_list_to_range, rescale_list_to_other

pd.options.mode.chained_assignment = 'raise'


class HklKeyRegistrar(type):
    """Metaclass for `HklKey`s which registers them if they define `name`."""
    REGISTRY = {}
    default: None
    dtype: np.dtype

    def __new__(mcs, name, bases, attrs):
        new_cls = type.__new__(mcs, name, bases, attrs)
        if hasattr(new_cls, 'name') and new_cls.name:
            mcs.REGISTRY[new_cls.name] = new_cls
        return new_cls

    @property
    def IMPERATIVES(self):  # noqa - capital letters to avoid attribute clash
        return [k for k, v in self.REGISTRY.items() if v.imperative]


class HklKey(metaclass=HklKeyRegistrar):
    """Base Class for every subsequent HklKey."""
    name: str = ''                  # if not empty, how key will be registered
    default = None                  # default to set if value was not provided
    dtype: str = ''                 # dtype to use when defining numpy array
    imperative: bool = False        # is key imperative for every table?
    reduce_behaviour: str = 'keep'  # 'add', 'average', 'discard' or 'keep' 1st
    type: type = type(None)         # python type used when defining instances


class HklKeyImperativeInt8(HklKey):
    default = 0
    dtype = np.int8
    imperative = True


class HklKeyAveragedFloat64(HklKey):
    default = 0.0
    dtype = np.float64
    reduce_behaviour = 'average'


class HklKeyKeptFloat64(HklKey):
    default = 0.0
    dtype = np.float64
    reduce_behaviour = 'keep'


class HklKeyIndexH(HklKeyImperativeInt8):
    """Reciprocal lattice Miller index h"""
    name = 'h'


class HklKeyIndexK(HklKeyImperativeInt8):
    """Reciprocal lattice Miller index k"""
    name = 'k'


class HklKeyIndexL(HklKeyImperativeInt8):
    """Reciprocal lattice Miller index l"""
    name = 'l'


class HklKeyStructureFactor(HklKeyAveragedFloat64):
    """Crystallographic structure factor F"""
    name = 'F'
    default = 1.0


class HklKeyIntensity(HklKeyAveragedFloat64):
    """Crystallographic intensity I_obs"""
    name = 'I'
    default = 1.0


class HklKeyIntensityCalculated(HklKeyAveragedFloat64):
    """Crystallographic calculated intensity I_calc"""
    name = 'Ic'
    default = 1.0


class HklKeyIntensityUncertainty(HklKeyAveragedFloat64):
    """Uncertainty of intensity I determination"""
    name = 'si'


class HklKeyStructureFactorUncertainty(HklKeyAveragedFloat64):
    """Uncertainty of structure factor F determination"""
    name = 'sf'


class HklKeyStructureFactorToUncertaintyRatio(HklKeyAveragedFloat64):
    """Structure factor to its uncertainty ratio"""
    name = 'u'


class HklKeyBatchNumber(HklKey):
    """Batch or run number"""
    name = 'b'
    default = 0
    dtype = np.int16
    reduce_behaviour = 'discard'


class HklKeyCrystalNumber(HklKey):
    """Crystal domain or twin number"""
    name = 'c'
    default = 0
    dtype = np.int16
    reduce_behaviour = 'discard'


class HklKeyMultiplicity(HklKey):
    """Multiplicity i.e. how many observation contributed to this reflection"""
    name = 'm'
    default = 1
    dtype = np.int16
    imperative = True
    reduce_behaviour = 'add'


class HklKeyWavelength(HklKey):
    """Wavelength expressed in Angstrom"""
    name = 'la'
    default = 0.0
    dtype = np.float64
    reduce_behaviour = 'discard'


class HklKeyPhase(HklKeyAveragedFloat64):
    """Reflection phase expressed in radians"""
    name = 'ph'
    default = 0.0


class HklKeyRadius(HklKey):
    """Distance from 000 node, i.e. double the sin(th)/la resolution in A^-1"""
    name = 'r'
    dtype = np.float64


class HklKeyTransmissionPathLength(HklKeyKeptFloat64):
    """Absorption-weighted transmission path length in centimeters"""
    name = 't'


class HklKeyCosineU1(HklKeyKeptFloat64):
    """Direction cosines of a vector (unused, see XD manual for reference)"""
    name = 'u1'


class HklKeyCosineU2(HklKeyKeptFloat64):
    """Direction cosines of a vector (unused, see XD manual for reference)"""
    name = 'u2'


class HklKeyCosineU3(HklKeyKeptFloat64):
    """Direction cosines of a vector (unused, see XD manual for reference)"""
    name = 'u3'


class HklKeyCosineV1(HklKeyKeptFloat64):
    """Direction cosines of a vector (unused, see XD manual for reference)"""
    name = 'v1'


class HklKeyCosineV2(HklKeyKeptFloat64):
    """Direction cosines of a vector (unused, see XD manual for reference)"""
    name = 'v2'


class HklKeyCosineV3(HklKeyKeptFloat64):
    """Direction cosines of a vector (unused, see XD manual for reference)"""
    name = 'v3'


class HklKeyCoordinateX(HklKeyKeptFloat64):
    """Reciprocal space coordinate x*"""
    name = 'x'
    dtype = np.float32


class HklKeyCoordinateY(HklKeyKeptFloat64):
    """Reciprocal space coordinate x*"""
    name = 'y'
    dtype = np.float32


class HklKeyCoordinateZ(HklKeyKeptFloat64):
    """Reciprocal space coordinate x*"""
    name = 'z'
    dtype = np.float32


class HklKeyZeta(HklKeyAveragedFloat64):
    """weighted diff. in observed I - calculated Ic intensity"""
    name = 'ze'


class HklKeyZetaSquared(HklKeyAveragedFloat64):
    """weighted diff. in observed I - calculated Ic intensity, squared"""
    name = 'ze2'


class HklKeyCalculatedIntensityToSigma(HklKeyAveragedFloat64):
    """Calculated intensity of reflection divided by experimental sigma"""
    name = 'Icsi'
    default = 1.0


class HklKeyObservedIntensityToSigma(HklKeyAveragedFloat64):
    """Observed intensity of reflection divided by experimental sigma"""
    name = 'Iosi'
    default = 1.0


class HklKeyEquivalenceCode(HklKey):
    """Integer unique for each set of symmetrically equivalent reflections"""
    name = 'equiv'
    default = 0
    dtype = np.int64


class HklKeyDummy(HklKey):
    """Dummy column for loading or holding irrelevant string data"""
    name = 'None'
    default = ''
    dtype = np.str_


class HklFrame(BaseFrame):
    """
    A master object which manages single-crystal diffraction files.
    It utilises other `Hkl*` classes to import, store, manipulate and output
    information about single-crystal diffraction patterns.

    HklFrame acts as a container which stores
    the diffraction data (Pandas dataframe, :attr:`table`)
    and elementary crystal cell data (:class:`hikari.dataframes.Base`).
    Demanding methods belonging to this class are vectorized,
    providing relatively satisfactory performance and high memory capacity.
    HklFrame methods are designed to work in-place, so the work strategy
    is to create a new instance of HklFrame for each reflection dataset,
    manipulate it using methods, eg. :func:`merge` or :func:`trim`, and
    :func:`copy` to other object or output using :func:`write` if needed.

    The HklFrame always initiates empty and does not accept any arguments.
    Some magic methods, such as :func:`__len__` and :func:`__add__`
    are defined and describe/operate on the :attr:`frame`.
    """

    HKL_LIMIT = 127
    """Highest absolute value of h, k or l index, which can be
    interpreted correctly by current version of the software."""

    def __init__(self):
        """HklFrame constructor"""
        super().__init__()

        self.__la = 0.71069
        """Wavelength of radiation used in experiment."""

        self.table = pd.DataFrame()
        """
        Pandas dataframe containing diffraction data information.
        Each row represents one reflection observation,
        while each column has one piece of information about the reflections.
        For a list of available keys, see :class:`HklKeys`,
        whose instance is used to menage the keys of this table.
        """

    def __add__(self, other):
        """
        :param other: HklFrame to be added to data
        :type other: HklFrame
        :return: concatenated :attr:`table` dataframes with metadata from first
        :rtype: HklFrame
        """
        _copied = self.copy()
        _copied.table = pd.concat([self.table, other.table], ignore_index=True)
        return _copied

    def __len__(self):
        """
        :return: Number of rows (individual reflections) in `self.data`
        :rtype: int
        """
        return self.table.shape[0]

    def __str__(self):
        """
        :return: Human-readable representation of `self.data`
        :rtype: str
        """
        return self.table.__str__()

    @property
    def la(self):
        """
        Wavelength of radiation used in the diffraction experiment.
        Can be set using popular abbreviations such as "MoKa" or "CuKb",
        where *a* and *b* stand for *alpha* and *beta*.
        Implemented cathode materials include:
        "Ag", "Co", "Cr", "Cu", "Fe", "Mn", "Mo", "Ni", "Pd", "Rh", "Ti", "Zn"
        and have been imported from International Tables
        of Crystallography, Volume C, Table 4.2.4.1, 3rd Edition.

        :return: wavelength of radiation used in experiment
        :rtype: float
        """
        return self.__la

    @la.setter
    def la(self, wavelength):
        try:
            self.__la = characteristic_radiation[wavelength[:4].lower()]
        except TypeError:
            self.__la = float(wavelength)

    @property
    def r_lim(self):
        """
        :return: Radius of limiting sphere in A^-1 calculated as 2/:attr:`la`
        :rtype: float
        """
        return 2.0 / self.la

    def _in_dacs(self, opening_angle, vectors):
        oa = np.deg2rad(opening_angle)
        xyz = self.table.loc[:, ('x', 'y', 'z')].to_numpy()
        r = self.table.loc[:, 'r'].to_numpy()
        v = np.array(vectors)
        v = (v.T / lin.norm(v, axis=1)).T              # normalise vectors v
        m1 = np.matmul(v, xyz.T)                       # dist from dac plane p
        phi = np.abs(np.arcsin((m1 / r).clip(-1, 1)))  # angle <(plane p, v)
        lim = self.r_lim * np.sin(oa - phi)            # True if in dac
        return r[None, :] < lim

    def dac_trim(self, opening_angle: float = 35.0, vector=None):
        r"""
        Remove reflections outside the opening_angle DAC-accessible volume.
        Sample/DAC orientation can be supplied either via specifying crystal
        orientation in :class:`hikari.dataframes.BaseFrame`, in
        :attr:`orientation` or providing a xyz\* *vector* perpendicular to the
        dac-accessible disc. For further details, see `*Tchoń & Makal, IUCrJ
        8, 1006-1017 (2021)* <https://doi.org/10.1107/s2052252521009532>`_.

        :param opening_angle: DAC single opening angle in degrees, default 35.
        :type opening_angle: float
        :param vector: Provides information about orientation of crystal
          relative to DAC. If None, :attr:`orientation` is used instead.
        :type vector: tuple[float]
        :return: HklFrame containing only reflections in dac-accessible region.
        :rtype: HklFrame
        """
        if vector is None:
            l_v = np.array((1.0, 0.0, 0.0))           # vector parallel to beam
            hkl = lin.inv(self.orientation) @ l_v     # calculate hkl vector
            v = hkl @ self.A_r                        # calculate xyz* vector
        else:
            v = vector
        in_dac = self._in_dacs(opening_angle, vectors=np.array([v, ]))[0]
        self.table = self.table[in_dac]
        self.table.reset_index(drop=True, inplace=True)

    def dacs_count(self, opening_angle: float = 35.0,
                   vectors: np.ndarray = np.array((1, 0, 0))):
        """
        Count unique dac-accessible reflections for n crystals placed such that
        vector n is perpendicular to diamond. For details see :meth:`dac_trim`.

        :param opening_angle: DAC single opening angle in degrees, default 35.
        :type opening_angle: float
        :param vectors: Array with rotational axes of available DAC-discs.
        :type vectors: np.array
        :return: Array with numbers of unique reflns in DAC-accessible region.
        :rtype: np.array
        """
        memory_estimate = 26 * len(self) * len(vectors)  # estimate memory use
        cycles_needed = -(memory_estimate // -hikari.MEMORY_SIZE)  # and split
        if cycles_needed > 1:
            vectors_split = np.array_split(vectors, cycles_needed)
            return np.hstack([self.dacs_count(opening_angle, vectors=v)
                              for v in vectors_split])
        else:
            in_dac = self._in_dacs(opening_angle, vectors)
            return np.array([self.table.loc[in_dac[n, :], 'equiv'].nunique()
                             for n in range(vectors.shape[0])])

    def copy(self):
        """
        :return: An exact deep copy of this HklFrame.
        :rtype: HklFrame
        """
        return copy.deepcopy(self)

    def extinct(self, space_group: Group = SG['P1']):
        """
        Removes from dataframe reflections which should be extinct based on
        space :class:`hikari.symmetry.group.Group`. For ref. see ITC-A12.3.5.

        :param space_group: Space group used to extinct the reflections.
        :type space_group: hikari.symmetry.group.Group
        """
        hkls = self.table.loc[:, ['h', 'k', 'l']].to_numpy()
        extinct_flag_list = [o.extincts(hkls) for o in space_group.operations]
        extinct_flag_list_union = np.logical_or.reduce(extinct_flag_list)
        self.table = self.table[~extinct_flag_list_union]
        self.table.reset_index(drop=True, inplace=True)

    def find_equivalents(self, point_group: Group = PG['1']):
        """
        Assign each reflection its symmetry equivalence identifier and store
        it in the `hikari.dataframes.HklFrame.data['equiv']` column.
        The ID is an integer unique for each set of equivalent reflections.

        In order to provide an information about equivalence, a *point_group*
        of reciprocal space must be provided (default PG['1']). Point groups
        and their notation can be found in :mod:`hikari.symmetry` sub-package.

        :param point_group: Point group used to determine symmetry equivalence
        :type point_group: hikari.symmetry.Group
        """
        inc = 10 ** (int(np.log10(self.HKL_LIMIT)) + 2)
        equiv_dtype = HklKey.REGISTRY['equiv'].dtype
        self.table.reset_index(drop=True, inplace=True)
        self.table['equiv'] = -inc**3
        _hkl_matrix = self.table.loc[:, ('h', 'k', 'l')].to_numpy()
        for op in point_group.operations:
            _new_hkl_matrix = op.transform(_hkl_matrix).astype(equiv_dtype)
            new_equiv = _new_hkl_matrix @ np.array([inc ** 2, inc, 1])
            _to_update = self.table['equiv'] < new_equiv
            self.table.loc[_to_update, 'equiv'] = new_equiv[_to_update]

    def from_dict(self, dictionary: dict):
        """
        Construct the `self.data` using information stored in dictionary.
        The dictionary keys must be valid strings, see :class:`HklKeys` for
        a list of valid keys. The dictionary values must be iterable of equal
        size, preferably `numpy.ndarray`.

        :param dictionary: Dictionary with "key - iterable of values" pairs.
        :type dictionary: Dict[str, numpy.ndarray]
        """
        df = pd.DataFrame()
        for key, value in dictionary.items():
            typ = HklKey.REGISTRY[key].dtype
            df[key] = pd.Series(value, dtype=typ, name=key)
        self.table = df[(df['h'] != 0) | (df['k'] != 0) | (df['l'] != 0)].copy()
        if not('x' in self.table.columns):
            self.place()
        self._recalculate_structure_factors_and_intensities()

    def fill(self, radius: float = 2.0) -> None:
        """
        Fill dataframe with all reflections within *radius* from space origin.

        :param radius: Maximum distance from the reciprocal space origin
            to placed reflection (in reciprocal Angstrom).
        """
        hkl_ratios = radius / np.array([self.a_r, self.b_r, self.c_r])
        hkl_limits = np.ceil(hkl_ratios).astype(np.int16)

        def hkl_walls(h, k, l_):
            hw = np.mgrid[h:h:1j, -k:k:2j*k+1j, -l_:l_:2j*l_+1j].reshape(3, -1)
            kw = np.mgrid[-h:h:2j*h+1j, k:k:1j, -l_:l_:2j*l_+1j].reshape(3, -1)
            lw = np.mgrid[-h:h:2j*h+1j, -k:k:2j*k+1j, l_:l_:1j].reshape(3, -1)
            return hw.T, kw.T, lw.T

        for _ in range(self.HKL_LIMIT):
            limits_too_small = [any(lin.norm(hkls @ self.A_r, axis=1) <= radius)
                                for hkls in hkl_walls(*hkl_limits)]
            if any(limits_too_small):
                hkl_limits += limits_too_small
            else:
                hkl_limits -= 1
                break

        if any(hkl_limits > self.HKL_LIMIT):
            msg = 'Attempting to use hkl indices {} above HKL_LIMIT of {}'
            raise ValueError(msg.format(hkl_limits, self.HKL_LIMIT))

        hkls = np.indices(2 * hkl_limits + 1, np.int16).reshape(3, -1).T \
            - hkl_limits
        _h, _k, _l = hkls[lin.norm(hkls @ self.A_r, axis=1) <= radius].T
        ones = np.ones_like(np.array(_h)[0])
        self.from_dict({'h': np.squeeze(_h), 'k': np.squeeze(_k),
                        'l': np.squeeze(_l), 'I': ones, 'si': ones, 'm': ones})

    def stats(self, bins: int = 10, space_group: Group = SG['P1']):
        """
        Returns completeness, redundancy, number of all, unique & theoretically
        possible reflections within equal-volume `bins` in given `space group`.

        :param bins: Number of equal-volume bins to divide the data into.
        :type bins: int
        :param space_group: Group used to calculate equivalence and extinctions
        :type space_group: hikari.symmetry.Group
        :return: String containing table with stats as a function of resolution
        :rtype: str
        """

        point_group = space_group.reciprocate()

        hkl_base = self.copy()
        hkl_base.extinct(space_group)
        hkl_base.find_equivalents(point_group)
        hkl_base.table['_i_to_si'] = hkl_base.table['I'] / hkl_base.table['si']

        hkl_full = self.copy()
        hkl_full.fill(radius=max(self.table['r']))
        hkl_full.merge(point_group)
        hkl_full.extinct(space_group)
        hkl_full.find_equivalents(point_group)

        def group_by_resolution(hkl_):
            bin_limits = cubespace(0.0, max(self.table['r']), num=bins + 1)
            return hkl_.table.groupby(by=pd.cut(hkl_.table['r'], bin_limits), observed=False)

        grouped_base = group_by_resolution(hkl_base)
        grouped_full = group_by_resolution(hkl_full)

        observed = grouped_base.size()
        independent = grouped_base['equiv'].nunique()
        theory = grouped_full['equiv'].nunique()
        cpl = independent.div(theory)
        red = observed.div(independent)
        i2si = grouped_base['_i_to_si'].mean()
        out = pd.concat([observed, independent, theory, i2si, cpl, red], axis=1)
        out.columns = ['Obser', 'Indep', 'Theory', 'I/si(I)', 'Cplt', 'Red.']
        return out  # use .reset_index().to_string(index=False) to flatten

    def merge(self, point_group=PG['1']):
        """
        Average down each set of redundant reflections present in the table,
        to one reflection. The redundancy is determined using the
        :meth:`find_equivalents` method with appropriate point group. Thus,
        the merging can be used in different ways depending on point group:

        - For PG['1'], only reflections with exactly the same h, k, l indices
          will be merged. Resulting dataframe will not contain any duplicates.

        - For PG['-1'] reflections with the same h, k and l as well as their
          Friedel pairs will be merged together to one reflection.

        - For PG['mmm'] all equivalent reflections of "mmm" point group will be
          merged. Since "mmm" is centrosymmetric, Friedel pairs will be merged.

        - For PG['mm2'] symmetry-equivalent reflections within the "mmm"
          point group will be merged, but the Friedel pairs will be preserved.

        The procedure will have a different effect on different dataframe keys,
        depending on their "reduce_behaviour" specified in :class:`HklKeys`.
        Fixed parameters *h, k, l, x, y, z, r* and *equiv* will be preserved;
        Floating points such as intensity *I*, structure factor *F* and their
        uncertainties *si* and *sf* will be averaged using arithmetic mean;
        Multiplicity *m* will be summed; Other parameters which would lose
        their meaning such as batch number *b* will be discarded.

        The merging inevitably removes some information from the dataframe,
        but it can be necessary for some operations. For example, the drawing
        procedures work faster and provide clearer image if multiple points
        occupying the same position in space are reduced to one instance.

        :param point_group: Point Group used to determine symmetry equivalence
        :type point_group: hikari.symmetry.Group
        """
        self.find_equivalents(point_group=point_group)
        # group the dataframe and obtain all existing keys
        grouped = self.table.groupby('equiv')
        grouped_first = grouped.first().reset_index()
        grouped_mean = grouped.mean().reset_index()
        grouped_sum = grouped.sum().reset_index()
        # for each key apply a necessary reduce operation and add it to data
        data = dict()
        for key in self.table.keys():
            if HklKey.REGISTRY[key].reduce_behaviour == 'keep':
                data[key] = grouped_first[key]
            elif HklKey.REGISTRY[key].reduce_behaviour == 'add':
                data[key] = grouped_sum[key]
            elif HklKey.REGISTRY[key].reduce_behaviour == 'average':
                data[key] = grouped_mean[key]
        self.from_dict(data)

    def place(self):
        """
        Assign reflections their positions in reciprocal space ("x", "y", "z")
        and calculate their distance from origin ("r") in reciprocal Angstrom.
        Save four new keys and their values into the dataframe.
        """
        hkl = self.table.loc[:, ('h', 'k', 'l')].to_numpy()
        xyz = hkl @ self.A_r
        self.table.loc[:, 'x'] = xyz[:, 0]
        self.table.loc[:, 'y'] = xyz[:, 1]
        self.table.loc[:, 'z'] = xyz[:, 2]
        self.table.loc[:, 'r'] = lin.norm(xyz, axis=1)

    def calculate_fcf_statistics(self):
        """
        Calculate values of zeta (I - Ic) / si on other stats based on contents
        of fcf files. Save new key and its values into the dataframe.
        """
        ze = (self.table['I'] - self.table['Ic']) / self.table['si']
        self.table['ze'] = ze
        self.table['ze2'] = ze ** 2
        self.table['Iosi'] = self.table['I'] / self.table['si']
        self.table['Icsi'] = self.table['Ic'] / self.table['si']

    def read(self, hkl_path, hkl_format='shelx_4'):
        """
        Read the contents of .hkl file as specified by path and format,
        and store them in the pandas dataframe in `self.data`.
        For a list of all available .hkl formats,
        please refer to :attr:`hikari.dataframes.HklIo.format`.

        :param hkl_path: Absolute or relative path to the .hkl file.
        :type hkl_path: str
        :param hkl_format: Format of provided .hkl file.
        :type hkl_format: union[int, str, dict]
        """
        reader = HklReader(hkl_file_path=hkl_path, hkl_file_format=hkl_format)
        dict_of_data = reader.read()
        forgotten_keys = [k for k in HklKey.IMPERATIVES
                          if k not in dict_of_data.keys()]
        for key in forgotten_keys:
            default = HklKey.REGISTRY[key].default
            length_of_data = max([len(v) for v in dict_of_data.values()])
            dict_of_data[key] = [default] * length_of_data
        self.from_dict(dict_of_data)

    def _recalculate_structure_factors_and_intensities(self):
        """
        Calculate 'I' and 'si' or 'F' and 'sf', depending on which are missing.
        """
        if 'I' in self.table.keys() and 'si' not in self.table.keys():
            raise KeyError('Intensities "I" are defined, but "si" not.')
        if 'F' in self.table.keys() and 'sf' not in self.table.keys():
            raise KeyError('Structure factors "F" are defined, but "sf" not.')
        if all(('I' in self.table.keys(), 'si' in self.table.keys(),
                'F' not in self.table.keys(), 'sf' not in self.table.keys())):
            self._recalculate_structure_factors_from_intensities()
        if all(('F' in self.table.keys(), 'sf' in self.table.keys(),
                'I' not in self.table.keys(), 'si' not in self.table.keys())):
            self._recalculate_intensities_from_structure_factors()

    def _recalculate_structure_factors_from_intensities(self):
        r"""
        Recalculate the structure factor F and its uncertainty sf.

        Structure factor is calculated as follows:
        *F = signum(I) \* sqrt(abs(I))*.

        Structure factor's uncertainty is calculated as follows:
        *sf = si / (2 \* sqrt(abs(I)))*.
        """
        new_data = copy.deepcopy(self.table)
        signum_of_i = new_data['I'].copy()
        signum_of_i[signum_of_i > 0] = 1
        signum_of_i[signum_of_i < 0] = -1
        absolute_sqrt_of_i = abs(new_data["I"]) ** 0.5
        new_data['F'] = signum_of_i * absolute_sqrt_of_i
        new_data['sf'] = new_data["si"] / (2 * absolute_sqrt_of_i)
        self.table = new_data

    def _recalculate_intensities_from_structure_factors(self):
        r"""
        Recalculate the intensity I and its uncertainty si.

        Intensity is calculated as follows:
        *I = signum(F) \* F \*\* 2*.

        Intensity's uncertainty is calculated as follows:
        *si = 2 \* sf \* abs(F)*.
        """
        new_data = copy.deepcopy(self.table)
        signum_of_f = new_data['F'].copy()
        signum_of_f[signum_of_f > 0] = 1
        signum_of_f[signum_of_f < 0] = -1
        new_data['I'] = signum_of_f * (abs(new_data['F']) ** 2)
        new_data['si'] = 2 * new_data["sf"] * abs(new_data["F"])
        self.table = new_data

    def transform(self, operations: Union[Iterable[np.ndarray], np.ndarray]):
        """
        Apply a symmetry operation or list of symmetry operations to transform
        the diffraction pattern.

        If one symmetry operation (3x3 or 4x4 numpy array) is provided,
        it effectively multiplies the hkl matrix by the operation matrix
        and accordingly alters the `self.data` dataframe. As a result,
        the length of `self.data` before and after transformation is the same.

        However, the function behaves slightly counter-intuitively
        if two or more operation matrices are provided. In such case
        the method applies the transformation procedure independently
        for each operation, and then *concatenates* resulting matrices.
        Resulting self.data is len(operations) times longer than the initial.

        The function can use 3x3 or larger (e.g. 4x4) matrices, as it selects
        only the upper-left 3x3 segment for the sake of calculations.
        Also, while reconstructing the symmetry of merged reflection file
        it is important to use all symmetry operations, not only generators.

        Single symmetry operations or their lists belonging to certain
        point groups can be imported from :py:mod:`hikari.symmetry` module.

        :param operations: Iterable of operation matrices to be applied
        :type operations: Union[Iterable[np.ndarray], np.ndarray]
        """

        def _make_a_list_of_operations():
            _ops = np.array(operations)
            if len(_ops.shape) == 1:
                raise ValueError('Operations should be a list of matrices')
            elif len(_ops.shape) == 2 and _ops.shape[0] == _ops.shape[1]:
                return [_ops]
            elif len(_ops.shape) == 3 and _ops.shape[1] == _ops.shape[2]:
                return operations
            else:
                raise ValueError('Operations should be a list of matrices')
        ops = _make_a_list_of_operations()

        def _build_transformed_dataframe():
            dataframes = list()
            for op in ops:
                df = copy.deepcopy(self.table)
                mat = op[0:3, 0:3]
                hkl = (mat @ self.table.loc[:, ['h', 'k', 'l']].to_numpy().T).T
                df['h'] = hkl[:, 0]
                df['k'] = hkl[:, 1]
                df['l'] = hkl[:, 2]
                dataframes.append(df)
            return pd.concat(dataframes, axis=0, ignore_index=True)
        self.table = _build_transformed_dataframe()
        self.place()

    def thin_out(self, target_cplt=1.0):
        """
        Randomly remove reflections from dataframe in order to decrease
        the completeness to *target_cplt* (relatively to initial completeness).

        :param target_cplt: Percentage of data not removed from dataframe
        :type target_cplt: float
        """
        assert 0.0 <= target_cplt <= 1.0, 'target_cplt outside the 0--1 range'
        number_to_delete = int((1 - target_cplt) * len(self))
        indices_to_delete = random.sample(range(0, len(self)), number_to_delete)
        self.table.drop(indices_to_delete, inplace=True)
        self.table.reset_index(drop=True, inplace=True)

    def to_res(self, path='hkl.res', colored='m'):
        """
        Export the reflection information from table to .res file,
        so that a software used to visualize .res files can be used
        to visualize a diffraction data in three dimensions.

        :param colored: Which key of dataframe should be visualized using color
        :type colored: str
        :param path: Absolute or relative path where the file should be saved
        :type path: str
        """
        HklToResConverter(self).convert(path)

    def trim(self, limit: float):
        """
        Remove reflections further than *limit* from reciprocal space origin.

        :param limit: Radius of the trimming sphere in reciprocal Angstrom
        :type limit: float
        """
        self.table = self.table.loc[self.table['r'] <= limit]

    def write(self, hkl_path, hkl_format='shelx_4'):
        """
        Write the contents of dataframe to a .hkl file using specified
        *path* and *format*.
        For a list of all available .hkl formats,
        please refer to :attr:`hikari.dataframes.HklIo.format`.

        :param hkl_path: Absolute or relative path to the .hkl file.
        :type hkl_path: str
        :param hkl_format: Desired format of .hkl file.
        :type hkl_format: union[int, str, dict]
        """
        writer = HklWriter(hkl_file_path=hkl_path, hkl_file_format=hkl_format)
        writer.write(hkl_data=self.table)


class HklIo:
    """
    A helper class supporting HklFrame. Manages reading and writing hkl files
    into and out of HklFrame's dataframe.
    """

    def __init__(self, hkl_file_path, hkl_file_format):
        self.use_separator = True
        self.file_path = make_abspath(hkl_file_path)
        self.formats_defined = hkl_formats
        self.formats_aliases = hkl_aliases
        self.__format = 'shelx_4'
        self.format = hkl_file_format

    def _build_line_formatter(self):
        """
        Prepare :meth:`line_formatter` to format hkl data while writing.

        :return: String for str.format() to format hkl data while writing.
        :rtype: str
        """
        formatter = str()
        for label, width in zip(self._format_dict['labels'],
                                self._format_dict['widths']):
            width = abs(width)
            text_width = abs(width) - int(self.use_separator)
            formatter += '{{{0}:>{1}.{2}}}'.format(label, width, text_width)
        formatter += '\n'
        self.__line_formatter = formatter

    @property
    def format(self):
        """
        Return a name of currently used hkl file format. Available file formats
        and their aliases are defined internally in .json files and
        have been presented in a table below:

        +----------+----------+------------------------+------+------+--------+
        | Name     | Aliases  | Contents               |Prefix|Suffix| Free   |
        |          |          | (format string)        |      |      | format |
        +==========+==========+========================+======+======+========+
        | free_2   |          | h -4 k -4 l -4 I -8    | NO   | YES  | YES    |
        |          |          | si -8 b -4 la -8       |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | free_3   |          | h -4 k -4 l -4         | NO   | YES  | YES    |
        |          |          | F -8 sf -8 b -4        |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | free_4   |          | h -4 k -4 l -4         | NO   | YES  | YES    |
        |          |          | I -8 si -8 b -4        |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | free_40  | free     | h -4 k -4 l -4         | NO   | YES  | YES    |
        |          |          | I -8 si -8             |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | free_5   |          | h -4 k -4 l -4         | NO   | YES  | YES    |
        |          |          | I -8 si -8 c -4        |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | free_6   |          | h -4 k -4 l -4         | NO   | YES  | YES    |
        |          |          | I -8 si -8 m -4        |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | m80      |          | h 4 k 4 l 4 b 4 F 12   | YES  | YES  | NO     |
        |          |          | None 132               | (b)  | (b)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_2  | 2        | h 4 k 4 l 4 I 8 si 8   | NO   | YES  | NO     |
        |          |          | b 4 la 8               |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_3  | 3        | h 4 k 4 l 4            | NO   | YES  | NO     |
        |          |          | F 8 sf 8 b 4           |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_4  | 4        | h 4 k 4 l 4            | NO   | YES  | NO     |
        |          |          | I 8 si 8 b 4           |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_40 | 40       | h 4 k 4 l 4            | NO   | YES  | NO     |
        |          |          | I 8 si 8               |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_5  | 5        | h 4 k 4 l 4            | NO   | YES  | NO     |
        |          |          | I 8 si 8 c 4           |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_6  | 6        | h 4 k 4 l 4            | NO   | YES  | NO     |
        |          |          | I 8 si 8 m 4           |      | (a)  |        |
        +----------+----------+------------------------+------+------+--------+
        | shelx_fcf| fcf      | h 4 k 4 l 4            | YES  | NO   | NO     |
        |          |          | Ic 14 I 14 si 13       | (*)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | tonto_F  |          | h -4 k -4 l -4         | YES  | YES  | YES    |
        |          |          | F -8 sf -8             | (c)  | (c)  |        |
        +----------+----------+------------------------+------+------+--------+
        | tonto_I  | tonto    | h -4 k -4 l -4         | YES  | YES  | YES    |
        |          |          | I -8 si -8             | (c)  | (c)  |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_F6    |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | F -13 sf -13           | (d)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_F7    |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | F -13 sf -13 t -10     | (d)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_F-7   |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | F -13 sf -13 ph -10    | (d)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_F13   |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | F -13 sf -13 t -10     | (d)  |      |        |
        |          |          | u1 -10 u2 -10 u3 -10   |      |      |        |
        |          |          | v1 -10 v2 -10 v3 -10   |      |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_I6    | xd       | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | I -13 si -13           | (d)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_I7    |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | I -13 si -13 t -10     | (d)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_I-7   |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | I -13 si -13 ph -10    | (d)  |      |        |
        +----------+----------+------------------------+------+------+--------+
        | xd_I13   |          | h -4 k -4 l -4 b -3    | YES  | NO   | YES    |
        |          |          | I -13 si -13 t -10     | (d)  |      |        |
        |          |          | u1 -10 u2 -10 u3 -10   |      |      |        |
        |          |          | v1 -10 v2 -10 v3 -10   |      |      |        |
        +----------+----------+------------------------+------+------+--------+
        | *custom* |          | custom string as above | NO   | NO   | if all |
        |          |          | with keys and widths   |      |      | widths |
        |          |          |                        |      |      | in     |
        |          |          |                        |      |      | format |
        |          |          |                        |      |      | are <0 |
        +----------+----------+------------------------+------+------+--------+

        Three different types of prefix / suffix are supported at the moment:

        - Suffix (a) is a zero-line: a shelx ending line with h = k = l = 0,

        - Prefix and suffix (b) express a superflip-style block/start end,

        - Prefix and suffix (c) are tonto-characteristic beginning/end of file,

        - Prefix (d) is an xd-characteristic line with info about file content.

        Pre/suffixes denoted with (*) are not supported in terms of writing.
        A custom hkl file format can be defined by providing
        a *format string* instead of 'Name'.
        The string should look like the ones in column "contents".
        For the meaning of keys ('I', 'b', 'c' etc.),
        please refer to :class:`HklKeys`.

        :return: Returns a name of currently used format.
        :rtype: str
        """
        return self.__format

    @format.setter
    def format(self, new_format):
        if new_format in self.formats_defined.keys():
            self.__format = new_format
        elif new_format in self.formats_aliases.keys():
            self.__format = self.formats_aliases[new_format]
        else:
            raise KeyError('Unknown hkl format "{}" given'.format(new_format))
        self._build_line_formatter()

    @property
    def _format_dict(self):
        """A dictionary with details concerning current format."""
        return self.formats_defined[self.__format]

    @property
    def _line_formatter(self):
        """A string used by string.format() method to print hkl data."""
        return self.__line_formatter

    def _import_custom_format(self, custom_format_string):
        """
        Import format string such as 'h 4 k 4 l 4 I 8 si 8' as format 'custom'.

        :param custom_format_string: string containing alternating data labels
            and column widths (all negative if free format) separated using ' '
        :type custom_format_string: str
        :return: None
        """
        custom_format_list = custom_format_string.strip(' ').split(' ')
        custom_labels = custom_format_list[0::2]
        custom_widths = [int(width) for width in custom_format_list[1::2]]
        assert len(custom_labels) == len(custom_widths), "Bad custom format"
        self.formats_defined['custom'] = {'labels': custom_labels,
                                          'widths': custom_widths,
                                          'prefix': '',
                                          'suffix': ''}

    @property
    def is_current_format_free(self):
        """
        Return true if currently defined format is free,
        i.e. the columns are separated by whitespace.

        :return: True if all format widths are negative; False otherwise.
        :rtype: bool
        """
        return all(width < 0 for width in self._format_dict['widths'])


class HklReader(HklIo):
    """
    A helper class for HklFrame,
    Manages reading hkl files and importing data and keys from them
    """

    def __init__(self, hkl_file_path, hkl_file_format):
        super().__init__(hkl_file_path, hkl_file_format)

    def _parse_fixed_line(self, line):
        """
        Parse data from a line, where data from each *label* has fixed *width*.

        :param line: string to be parsed based on format dictionary.
        :type line: str
        :return: list of strings extracted from parsed line
        :rtype: list
        """
        slice_end = list(np.cumsum(self._format_dict['widths']))
        slice_beg = [0] + slice_end[:-1]
        try:
            parsed = [line[beg:end] for beg, end in zip(slice_beg, slice_end)]
            parsed = np.array(parsed)
            assert min([len(p) for p in parsed]) > 0
            assert len(parsed) == len(self._format_dict['widths'])
        except (ValueError, AssertionError):
            return None
        return parsed

    def _parse_free_line(self, line):
        """
        Parse data from line, where data from *labels* is separated with space.

        :param line: string to be parsed based on format dictionary.
        :type line: str
        :return: list of strings extracted from parsed line
        :rtype: list
        """
        parsed = np.array(line.strip().split())
        try:
            parsed.astype('float64')
            assert len(parsed) == len(self._format_dict['widths'])
        except (ValueError, AssertionError):
            return None
        return parsed

    def read(self):
        """
        Read the contents of file currently pointed by :attr:`hkl_file_path`
        and format :attr:`hkl_file_format` and return them to a dictionary.

        :return: A dictionary containing information read from .hkl file.
        :rtype: dict
        """
        if self.is_current_format_free:
            def parse_line(line):
                return self._parse_free_line(line)
        else:
            def parse_line(line):
                return self._parse_fixed_line(line)

        def read_file_to_list_of_data():
            list_of_reflections = list()
            with open(self.file_path, 'r') as hkl_file:
                for line in hkl_file.read().splitlines():
                    parsed_line = parse_line(line)
                    if parsed_line is None:
                        continue
                    list_of_reflections.append(parsed_line)
            return np.array(list_of_reflections).transpose()
        array_of_reflections = read_file_to_list_of_data()

        def build_dict_of_reflections(_hkl_array):
            dict_of_data = dict()
            for index, key in enumerate(self._format_dict['labels']):
                key_dtype = HklKey.REGISTRY[key].dtype
                dict_of_data[key] = _hkl_array[index].astype(key_dtype)
            return dict_of_data
        return build_dict_of_reflections(array_of_reflections)


class HklWriter(HklIo):
    """
    A helper class for HklFrame,
    Manages writing hkl files and exporting data to them
    """
    def __init__(self, hkl_file_path, hkl_file_format):
        super().__init__(hkl_file_path, hkl_file_format)

    def write(self, hkl_data):
        """
        Write data from pandas dataframe `hkl_data` to the file specified 
        at :attr:`hkl_file_path` of format :attr:`hkl_file_format`.
        
        :param hkl_data: Dataframe containing reflection information.
        :type hkl_data: pandas.dataframe
        """
        needed_data = hkl_data.loc[:, self._format_dict['labels']].astype(str)
        with open(self.file_path, 'w') as hkl_file:
            hkl_file.write(self._format_dict['prefix'])
            for row_tuple in needed_data.itertuples(index=False):
                row_dict = dict(zip(self._format_dict['labels'], row_tuple))
                hkl_file.write(self._line_formatter.format(**row_dict))
            hkl_file.write(self._format_dict['suffix'])


class HklToResConverter:
    """A class responsible for representing hkl data using .res format"""

    MIN_DISTANCE = 10.0
    ELEMENTS = chemical_elements[:100]
    MIN_U = 0.00001
    MAX_U = 4.99999

    def __init__(self, hkl_dataframe):
        self.df = hkl_dataframe

    @property
    def abc_scale_factor(self):
        return self.largest_absolute_hkl * self.MIN_DISTANCE \
               / min(self.df.a_r, self.df.b_r, self.df.c_r)

    @property
    def largest_absolute_hkl(self):
        return self.df.table[['h', 'k', 'l']].abs().max().max()

    @property
    def x(self):
        return self.df.table['h'] / self.largest_absolute_hkl

    @property
    def y(self):
        return self.df.table['k'] / self.largest_absolute_hkl

    @property
    def z(self):
        return self.df.table['l'] / self.largest_absolute_hkl

    @property
    def u(self):
        if 'F' in self.df.table.columns:
            return rescale_list_to_range(self.df.table['F'],
                                         (self.MIN_U, self.MAX_U))
        else:
            return [1.0] * len(self.df.table)

    @property
    def c(self):
        if 'm' in self.df.table.columns:
            return rescale_list_to_other(self.df.table['m'], self.ELEMENTS)
        else:
            return [self.ELEMENTS[0]] * len(self.df.table)

    @property
    def res_header(self):
        return "TITL Reflection visualisation\n" \
               "REM Special file to be used in mercury with hkl.msd style.\n" \
               "REM Reciprocal unit cell inflated by {sc} to prevent bonds\n" \
               "CELL {la:7f} {a:7f} {b:7f} {c:7f} {al:7f} {be:7f} {ga:7f}\n" \
               "LATT -1\n\n".format(sc=self.abc_scale_factor,
                                    la=self.df.la,
                                    a=self.df.a_r * self.abc_scale_factor,
                                    b=self.df.b_r * self.abc_scale_factor,
                                    c=self.df.c_r * self.abc_scale_factor,
                                    al=np.rad2deg(self.df.al_r),
                                    be=np.rad2deg(self.df.be_r),
                                    ga=np.rad2deg(self.df.ga_r))

    @property
    def atom_list(self):
        cols = [self.c, self.df.table['h'], self.df.table['k'],
                self.df.table['l'], self.x, self.y, self.z, self.u]
        return [list(row) for row in zip(*cols)]

    @staticmethod
    def res_line(_c, _h, _k, _l, _x, _y, _z, _u):
        """
        :return: res line filled based on input element, hkl, xyz and u.
        :rtype: str
        """
        label = f'{_c}({_h},{_k},{_l})'
        pos = f' {_x: 7.5f} {_y: 7.5f} {_z: 7.5f}'
        return f'{label:16}   1{pos} 11.0 {_u: 7.5f}\n'

    def convert(self, path='~'):
        with open(make_abspath(path), 'w') as res:
            res.write(self.res_header)
            for row in self.atom_list:
                res.write(self.res_line(*row))
        with open(make_abspath(path, '../hkl.msd'), 'w') as style:
            style.write(hkl_mercury_style)

    # TODO wrap table/data in getter/setter and make it automatically place,
    # TODO refresh, set keys etc.



from hikari.scripts.angular_explorer import angular_property_explorer_factory


def r1_map(a, b, c, al, be, ga,
           space_group='P1',
           axis='',
           fix_scale=False,
           histogram=True,
           opening_angle=35,
           orientation=None,
           path='~/sortav.lst',
           output_quality=3,
           resolution=1.2,
           wavelength='MoKa'):
    """
    Calculate and draw a r1 map for a given crystal in diamond anvil cell
    (DAC) with a given opening angle, as a function of crystal orientation.

    The script accepts unit cell & space group information, runs SHELXL,
    and reads value of R1 after single refinement. Results are logged into text
    files and drawn with gnuplot or matplotlib, depending on settings.

    For further detail concerning r1_map, its basis and uses, refer to
    :py:func:`hikari.scripts.potency_map`, as well as selected terminology
    described in `this paper <https://doi.org/10.1107/S2052252521009532>`_.
    """
    kwargs = locals()
    ape = angular_property_explorer_factory.create(prop='r1')
    ape.set_up(**kwargs)
    ape.explore()
    ape.write_hist_file()
    ape.draw_matplotlib_map()
    ape.draw_gnuplot_map()


if __name__ == '__main__':
    r1_map(5.64109, 5.64109, 5.64109, 90, 90, 90, space_group='Fm-3m',
           path='~/_/NaCl/NaCl.hkl', output_quality=5, wavelength='MoKa')
